# Copyright (C) 2016 MushMush Foundation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from os import path
from datetime import datetime

import unittest
import tempfile
import shutil
import json

from conpot.core.loggers.json_log import JsonLogger


class TestJsonLogger(unittest.TestCase):
    def setUp(self):
        self.logging_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.logging_dir)

    def test_log_event(self):
        filename = path.join(self.logging_dir, 'test.json')
        sensorid = 'default'
        public_ip = '0.0.0.0'
        timestamp = datetime.now()
        event_id = 1337
        src_ip = '127.0.0.1'
        src_port = '2048'
        data_type = 'unittest'
        request = 'ping'
        response = 'pong'

        json_logger = JsonLogger(filename, sensorid, public_ip)
        json_logger.log({
            'timestamp': timestamp,
            'id': event_id,
            'remote': (src_ip, src_port),
            'data_type': data_type,
            'data': {
                'request': request,
                'response': response
            },
        })

        with open(filename, 'r') as logfile:
            e = json.load(logfile)
            self.assertEqual(e['timestamp'], timestamp.isoformat())
            self.assertEqual(e['sensorid'], sensorid)
            self.assertEqual(e['id'], event_id)
            self.assertEqual(e['src_ip'], src_ip)
            self.assertEqual(e['src_port'], src_port)
            self.assertEqual(e['dst_ip'], public_ip)
            self.assertEqual(e['data_type'], data_type)
            self.assertEqual(e['request'], request)
            self.assertEqual(e['response'], response)
            self.assertEqual(e['event_type'], None)



# Copyright (C) 2017  Patrick Reichenberger (University of Passau) <patrick.reichenberger@t-online.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from lxml import etree
from conpot.protocols.IEC104.frames import *
import conpot.core as conpot_core
from conpot.protocols.IEC104.register import IEC104Register

logger = logging.getLogger(__name__)


# Manages the devices in a dictionary with key: address in 16_8 Bit format and value: register objects
class DeviceDataController(object):
    def __init__(self, template):
        # key: IEC104 address, value: register object
        self.registers = {}
        dom = etree.parse(template)
        categories = dom.xpath('//IEC104/categories/*')

        for category in categories:
            categ_id = int(category.attrib['id'])
            for register in category:
                address = register.attrib['name']
                splt_addr1, splt_addr2 = address.split("_")
                assert (0 <= int(splt_addr1) <= 65535 and 0 <= int(splt_addr2) <= 255),\
                    "Address %s not allowed. 0..65535_0..255" % address
                databuskey = register.xpath('./value/text()')[0]
                if register.get('rel'):
                    rel = register.attrib['rel']
                else:
                    rel = ""

                # checks if a value for that key exists in xml file
                try:
                    val = conpot_core.get_databus().get_value(databuskey)
                except AssertionError as err:
                    err.args = ("Key not found in key-value store",)
                    raise

                # simple data type checks
                assert (not(categ_id in (1, 2, 30, 45, 58) and val not in (0, 1))),\
                    "Value for obj %s not allowed with datatype %s" % (address, categ_id)
                assert (not(categ_id in (3, 4, 31, 46, 59) and val not in (0, 1, 2, 3))), \
                    "Value for obj %s not allowed with datatype %s" % (address, categ_id)
                assert (not(categ_id in (11, 12, 49, 62) and -32768 >= val >= 32767)), \
                    "Value for obj %s not allowed with datatype %s" % (address, categ_id)
                iec104_register = IEC104Register(categ_id, address, val, rel)
                assert address not in self.registers
                self.registers[address] = iec104_register
        self.check_registers()

    # Checks if relation (if stated) exists
    def check_registers(self):
        for elem in self.registers:
            rel = self.registers[elem].relation
            assert (not (rel != "" and rel not in self.registers)), "Relation object doesn't exist"

    # Returns the object with the obj_addr from the register dictionary
    def get_object_from_reg(self, obj_addr):
        address_structured = hex_in_addr(obj_addr)
        if address_structured in self.registers:
            return self.registers[address_structured]
        else:
            return None

    # Sets the value for an object in the register list
    def set_object_val(self, obj_addr, val):
        address_hex = hex(obj_addr)
        address_structured = hex_in_addr(address_hex)
        if address_structured in self.registers:
            self.registers[address_structured].set_val(val)

    def get_registers(self):
        return self.registers


# Builds response for a certain asdu type and returns list of responses with this type
def inro_response(sorted_reg, asdu_type):
    resp_list = []
    resp = i_frame() / asdu_head(SQ=0, COT=20)
    max_frame_size = conpot_core.get_databus().get_value('MaxFrameSize')
    counter = 0
    asdu_infobj_type = "asdu_infobj_"+str(asdu_type)
    calls_dict = {'asdu_infobj_1': asdu_infobj_1, 'asdu_infobj_3': asdu_infobj_3,
                  'asdu_infobj_5': asdu_infobj_5, 'asdu_infobj_7': asdu_infobj_7,
                  'asdu_infobj_9': asdu_infobj_9, 'asdu_infobj_11': asdu_infobj_11,
                  'asdu_infobj_13': asdu_infobj_13}
    call = calls_dict[asdu_infobj_type]
    for dev in sorted_reg:
        if dev[1].category_id == asdu_type:
            # 12 is length i_frame = 6 + length asdu_head = 6
            if counter >= int((max_frame_size - 12) / len(call())):
                resp_list.append(resp)
                counter = 0
                resp = i_frame() / asdu_head(SQ=0, COT=20)
            xaddr = addr_in_hex(dev[1].addr)

            add_info_obj = call(IOA=xaddr)  # SQ = 0
            val = dev[1].val
            if asdu_type == 1:
                add_info_obj.SIQ = SIQ(SPI=val)
                # Other possibility for allocation (certain value for whole field)
                # add_info_obj.SIQ = struct.pack("B", val)
            elif asdu_type == 3:
                add_info_obj.DIQ = DIQ(DPI=val)
            elif asdu_type == 5:
                add_info_obj.VTI = VTI(Value=val)
            elif asdu_type == 7:
                add_info_obj.BSI = val
            elif asdu_type == 9:
                add_info_obj.NVA = val
            elif asdu_type == 11:
                add_info_obj.SVA = val
            elif asdu_type == 13:
                add_info_obj.FPNumber = val
            resp /= add_info_obj
            counter += 1
            resp.NoO = counter
    if counter > 0:
        resp_list.append(resp)
    return resp_list


# Converts the address from number representation in 16_8 Bit String format with delimiter "_"
def hex_in_addr(hex_addr):
        hexa = "{0:#0{1}x}".format(hex_addr, 8)
        a1 = hexa[2:4]
        a2 = hexa[4:6]
        a3 = hexa[6:8]
        a32 = a3 + a2
        return str(int(a32, 16)) + '_' + str(int(a1, 16))


# Converts the address from 16_8 Bit String format with delimiter "_" in a number representation
def addr_in_hex(address):
        a1, a2 = address.split("_")
        hex1_temp = "{0:0{1}x}".format(int(a1), 4)
        hex1_1 = hex1_temp[0:2]
        hex1_2 = hex1_temp[2:4]
        hex1 = str(hex1_2) + str(hex1_1)
        hex2 = "{0:#0{1}x}".format(int(a2), 4)  # bec of '0x' length 4
        hex_string = str(hex2) + str(hex1)
        return int(hex_string, 16)



# Copyright (C) 2017  Yuru Shao <shaoyuru@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import socket
import cpppo
import contextlib
import time
import sys
import traceback

from lxml import etree
from cpppo.server import network
from cpppo.server.enip import logix
from cpppo.server.enip import parser
from cpppo.server.enip import device
from conpot.core.protocol_wrapper import conpot_protocol
logger = logging.getLogger(__name__)


class EnipConfig(object):
    """
    Configurations parsed from template
    """

    def __init__(self, template):
        self.template = template
        self.parse_template()

    class Tag(object):
        """
        Represents device tag setting parsed from template
        """
        def __init__(self, name, type, size, value, addr=None):
            self.name = name
            self.type = str(type).upper()
            self.size = size
            self.value = value
            self.addr = addr

    def parse_template(self):
        dom = etree.parse(self.template)
        self.server_addr    = dom.xpath('//enip/@host')[0]
        self.server_port    = int(dom.xpath('//enip/@port')[0])
        self.vendor_id      = int(dom.xpath('//enip/device_info/VendorId/text()')[0])
        self.device_type    = int(dom.xpath('//enip/device_info/DeviceType/text()')[0])
        self.product_rev    = int(dom.xpath('//enip/device_info/ProductRevision/text()')[0])
        self.product_code   = int(dom.xpath('//enip/device_info/ProductCode/text()')[0])
        self.product_name   = dom.xpath('//enip/device_info/ProductName/text()')[0]
        self.serial_number  = dom.xpath('//enip/device_info/SerialNumber/text()')[0]
        self.mode           = dom.xpath('//enip/mode/text()')[0]
        self.timeout        = float(dom.xpath('//enip/timeout/text()')[0])
        self.latency        = float(dom.xpath('//enip/latency/text()')[0])

        # parse device tags, these tags will be further processed by the ENIP server
        self.dtags = []
        for t in dom.xpath('//enip/tags/tag'):
            name = t.xpath('@name')[0]
            type = t.xpath('type/text()')[0]
            value = t.xpath('value/text()')[0]
            addr = t.xpath('addr/text()')[0]
            size = 1
            try:
                size = int(t.xpath('size/text()')[0])
            except:
                raise AssertionError("Invalid tag size: %r" % size)

            self.dtags.append(self.Tag(name, type, size, value, addr))


@conpot_protocol
class EnipServer(object):
    """
    Ethernet/IP server
    """

    def __init__(self, template, template_directory, args):
        self.config = EnipConfig(template)
        self.addr = self.config.server_addr
        self.port = self.config.server_port
        self.stopped = False
        self.connections = cpppo.dotdict()

        # all known tags
        self.tags = cpppo.dotdict()
        self.set_tags()

        logger.debug('ENIP server serial number: ' + self.config.serial_number)
        logger.debug('ENIP server product name: ' + self.config.product_name)

    def stats_for(self, peer):
        if peer is None:
            return None, None
        connkey = "%s_%d" % (peer[0].replace('.', '_'), peer[1])
        stats = self.connections.get(connkey)
        if stats is not None:
            return stats, connkey
        stats = cpppo.apidict(timeout=self.config.timeout)
        self.connections[connkey] = stats
        stats['requests'] = 0
        stats['received'] = 0
        stats['eof'] = False
        stats['interface'] = peer[0]
        stats['port'] = peer[1]
        return stats, connkey

    def handle(self, conn, address, enip_process=None, delay=None, **kwds):
        """
        Handle an incoming connection
        """
        name = "ENIP_%s" % (address[1] if address else "UDP")
        logger.debug("ENIP server %s begins serving client %s", name, address)

        tcp = (conn.family == socket.AF_INET and conn.type == socket.SOCK_STREAM)
        udp = (conn.family == socket.AF_INET and conn.type == socket.SOCK_DGRAM)

        if tcp:
            try:
                conn.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            except Exception as e:
                logger.error("%s unable to set TCP_NODELAY for client %r: %s", name, address, e)
            try:
                conn.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            except Exception as e:
                logger.error("%s unable to set SO_KEEPALIVE for client %r: %s", name, address, e)
            self.handle_tcp(conn, address, name=name, enip_process=enip_process, delay=delay, **kwds)
        elif udp:
            self.handle_udp(conn, name=name, enip_process=enip_process, **kwds)
        else:
            raise NotImplemented("Unknown socket protocol for EtherNet/IP CIP")

    def handle_tcp(self, conn, address, name, enip_process, delay=None, **kwds):
        """
        Handle a TCP client
        """
        source = cpppo.rememberable()
        with parser.enip_machine(name=name, context='enip') as machine:
            try:
                assert address, "EtherNet/IP CIP server for TCP/IP must be provided a peer address"
                stats, connkey = self.stats_for(address)
                while not stats.eof:
                    data = cpppo.dotdict()
                    source.forget()
                    # If no/partial EtherNet/IP header received, parsing will fail with a NonTerminal
                    # Exception (dfa exits in non-terminal state).  Build data.request.enip:
                    begun = cpppo.timer()
                    with contextlib.closing(machine.run(path='request', source=source, data=data)) as engine:
                        # PyPy compatibility; avoid deferred destruction of generators
                        for mch, sta in engine:
                            if sta is not None:
                                continue
                            # No more transitions available.  Wait for input.  EOF (b'') will lead to
                            # termination.  We will simulate non-blocking by looping on None (so we can
                            # check our options, in case they've been changed).  If we still have input
                            # available to process right now in 'source', we'll just check (0 timeout);
                            # otherwise, use the specified server.control.latency.
                            msg = None
                            while msg is None and not stats.eof:
                                wait = (kwds['server']['control']['latency'] if source.peek() is None else 0)
                                brx = cpppo.timer()
                                msg = network.recv(conn, timeout=wait)
                                now = cpppo.timer()
                                (logger.info if msg else logger.debug)(
                                    "Transaction receive after %7.3fs (%5s bytes in %7.3f/%7.3fs)",
                                    now - begun,
                                    len(msg) if msg is not None else "None",
                                    now - brx, wait)

                                # After each block of input (or None), check if the server is being
                                # signalled done/disabled; we need to shut down so signal eof.  Assumes
                                # that (shared) server.control.{done,disable} dotdict be in kwds.  We do
                                # *not* read using attributes here, to avoid reporting completion to
                                # external APIs (eg. web) awaiting reception of these signals.
                                if kwds['server']['control']['done'] or  kwds['server']['control']['disable']:
                                    logger.info("%s done, due to server done/disable", machine.name_centered())
                                    stats['eof'] = True
                                if msg is not None:
                                    stats['received'] += len(msg)
                                    stats['eof'] = stats['eof'] or not len(msg)
                                    if logger.getEffectiveLevel() <= logging.INFO:
                                        logger.info("%s recv: %5d: %s", machine.name_centered(), len(msg), cpppo.reprlib.repr(msg))
                                    source.chain(msg)
                                else:
                                    # No input.  If we have symbols available, no problem; continue.
                                    # This can occur if the state machine cannot make a transition on
                                    # the input symbol, indicating an unacceptable sentence for the
                                    # grammar.  If it cannot make progress, the machine will terminate
                                    # in a non-terminal state, rejecting the sentence.
                                    if source.peek() is not None:
                                        break
                                        # We're at a None (can't proceed), and no input is available.  This
                                        # is where we implement "Blocking"; just loop.

                    logger.info("Transaction parsed  after %7.3fs", cpppo.timer() - begun)
                    # Terminal state and EtherNet/IP header recognized, or clean EOF (no partial
                    # message); process and return response
                    if 'request' in data:
                        stats['requests'] += 1
                    try:
                        # enip_process must be able to handle no request (empty data), indicating the
                        # clean termination of the session if closed from this end (not required if
                        # enip_process returned False, indicating the connection was terminated by
                        # request.)
                        delayseconds = 0  # response delay (if any)
                        if enip_process(address, data=data, **kwds):
                            # Produce an EtherNet/IP response carrying the encapsulated response data.
                            # If no encapsulated data, ensure we also return a non-zero EtherNet/IP
                            # status.  A non-zero status indicates the end of the session.
                            assert 'response.enip' in data, "Expected EtherNet/IP response; none found"
                            if 'input' not in data.response.enip or not data.response.enip.input:
                                logger.warning("Expected EtherNet/IP response encapsulated message; none found")
                                assert data.response.enip.status, "If no/empty response payload, expected non-zero EtherNet/IP status"

                            rpy = parser.enip_encode(data.response.enip)
                            if logger.getEffectiveLevel() <= logging.INFO:
                                logger.info("%s send: %5d: %s %s", machine.name_centered(), len(rpy),
                                            cpppo.reprlib.repr(rpy), ("delay: %r" % delay) if delay else "")
                            if delay:
                                # A delay (anything with a delay.value attribute) == #[.#] (converible
                                # to float) is ok; may be changed via web interface.
                                try:
                                    delayseconds = float( delay.value if hasattr(delay, 'value') else delay)
                                    if delayseconds > 0:
                                        time.sleep(delayseconds)
                                except Exception as exc:
                                    logger.info( "Unable to delay; invalid seconds: %r", delay)
                            try:
                                conn.send(rpy)
                            except socket.error as exc:
                                logger.info("Session ended (client abandoned): %s", exc)
                                stats['eof'] = True
                            if data.response.enip.status:
                                logger.warning( "Session ended (server EtherNet/IP status: 0x%02x == %d)",
                                                data.response.enip.status, data.response.enip.status)
                                stats['eof'] = True
                        else:
                            # Session terminated.  No response, just drop connection.
                            if logger.getEffectiveLevel() <= logging.INFO:
                                logger.info("Session ended (client initiated): %s", parser.enip_format(data))
                            stats['eof'] = True
                        logger.info( "Transaction complete after %7.3fs (w/ %7.3fs delay)", cpppo.timer() - begun, delayseconds)
                    except:
                        logger.error("Failed request: %s", parser.enip_format(data))
                        enip_process(address, data=cpppo.dotdict())  # Terminate.
                        raise

                stats['processed'] = source.sent
            except:
                # Parsing failure.
                stats['processed'] = source.sent
                memory = bytes(bytearray(source.memory))
                pos = len(source.memory)
                future = bytes(bytearray(b for b in source))
                where = "at %d total bytes:\n%s\n%s (byte %d)" % (stats.processed, repr(memory + future), '-' * (len(repr(memory)) - 1) + '^', pos)
                logger.error("EtherNet/IP error %s\n\nFailed with exception:\n%s\n", where, ''.join(traceback.format_exception(*sys.exc_info())))
                raise
            finally:
                # Not strictly necessary to close (network.server_main will discard the socket,
                # implicitly closing it), but we'll do it explicitly here in case the thread doesn't die
                # for some other reason.  Clean up the connections entry for this connection address.
                self.connections.pop(connkey, None)
                logger.info( "%s done; processed %3d request%s over %5d byte%s/%5d received (%d connections remain)",
                    name, stats.requests, " " if stats.requests == 1  else "s",
                    stats.processed, " " if stats.processed == 1 else "s",
                    stats.received, len(self.connections))
                sys.stdout.flush()
                conn.close()

    def handle_udp(self, conn, name, enip_process, **kwds):
        """
        Process UDP packets from multiple clients
        """
        with parser.enip_machine(name=name, context='enip') as machine:
            while not kwds['server']['control']['done'] and not kwds['server']['control']['disable']:
                try:
                    source = cpppo.rememberable()
                    data = cpppo.dotdict()

                    # If no/partial EtherNet/IP header received, parsing will fail with a NonTerminal
                    # Exception (dfa exits in non-terminal state).  Build data.request.enip:
                    begun = cpppo.timer()  # waiting for next transaction
                    addr, stats = None, None
                    with contextlib.closing(machine.run(
                            path='request', source=source, data=data)) as engine:
                        # PyPy compatibility; avoid deferred destruction of generators
                        for mch, sta in engine:
                            if sta is not None:
                                # No more transitions available.  Wait for input.
                                continue
                            assert not addr, "Incomplete UDP request from client %r" % (addr)
                            msg = None
                            while msg is None:
                                # For UDP, we'll allow no input only at the start of a new request parse
                                # (addr is None); anything else will be considered a failed request Back
                                # to the trough for more symbols, after having already received a packet
                                # from a peer?  No go!
                                wait = (kwds['server']['control']['latency']
                                        if source.peek() is None else 0)
                                brx = cpppo.timer()
                                msg, frm = network.recvfrom(conn, timeout=wait)
                                now = cpppo.timer()
                                (logger.info if msg else logger.debug)(
                                    "Transaction receive after %7.3fs (%5s bytes in %7.3f/%7.3fs): %r",
                                    now - begun, len(msg) if msg is not None else "None",
                                    now - brx, wait, self.stats_for(frm)[0])
                                # If we're at a None (can't proceed), and we haven't yet received input,
                                # then this is where we implement "Blocking"; we just loop for input.

                            # We have received exactly one packet from an identified peer!
                            begun = now
                            addr = frm
                            stats, _ = self.stats_for(addr)
                            # For UDP, we don't ever receive incoming EOF, or set stats['eof'].
                            # However, we can respond to a manual eof (eg. from web interface) by
                            # ignoring the peer's packets.
                            assert stats and not stats.get('eof'), \
                                "Ignoring UDP request from client %r: %r" % (addr, msg)
                            stats['received'] += len(msg)
                            logger.debug("%s recv: %5d: %s", machine.name_centered(), len(msg), cpppo.reprlib.repr(msg))
                            source.chain(msg)

                    # Terminal state and EtherNet/IP header recognized; process and return response
                    assert stats
                    if 'request' in data:
                        stats['requests'] += 1
                    # enip_process must be able to handle no request (empty data), indicating the
                    # clean termination of the session if closed from this end (not required if
                    # enip_process returned False, indicating the connection was terminated by
                    # request.)
                    if enip_process(addr, data=data, **kwds):
                        # Produce an EtherNet/IP response carrying the encapsulated response data.
                        # If no encapsulated data, ensure we also return a non-zero EtherNet/IP
                        # status.  A non-zero status indicates the end of the session.
                        assert 'response.enip' in data, "Expected EtherNet/IP response; none found"
                        if 'input' not in data.response.enip or not data.response.enip.input:
                            logger.warning("Expected EtherNet/IP response encapsulated message; none found")
                            assert data.response.enip.status, "If no/empty response payload, expected non-zero EtherNet/IP status"

                        rpy = parser.enip_encode(data.response.enip)
                        logger.debug("%s send: %5d: %s", machine.name_centered(),
                                       len(rpy), cpppo.reprlib.repr(rpy))
                        conn.sendto(rpy, addr)

                    logger.debug("Transaction complete after %7.3fs", cpppo.timer() - begun)

                    stats['processed'] = source.sent
                except:
                    # Parsing failure.  Suck out some remaining input to give us some context, but don't re-raise
                    if stats:
                        stats['processed'] = source.sent
                    memory = bytes(bytearray(source.memory))
                    pos = len(source.memory)
                    future = bytes(bytearray(b for b in source))
                    where = "at %d total bytes:\n%s\n%s (byte %d)" % (
                        stats.get('processed', 0) if stats else 0,
                        repr(memory + future), '-' * (len(repr(memory)) - 1) + '^', pos)
                    logger.error("Client %r EtherNet/IP error %s\n\nFailed with exception:\n%s\n", addr, where, ''.join(
                        traceback.format_exception(*sys.exc_info())))

    def set_tags(self):
        typenames = {
            "BOOL":    (parser.BOOL,    0,   lambda v: bool(v)),
            "INT":     (parser.INT,     0,   lambda v: int(v)),
            "DINT":    (parser.DINT,    0,   lambda v: int(v)),
            "SINT":    (parser.SINT,    0,   lambda v: int(v)),
            "REAL":    (parser.REAL,    0.0, lambda v: float(v)),
            "SSTRING": (parser.SSTRING, '',  lambda v: str(v)),
            "STRING":  (parser.STRING,  '',  lambda v: str(v)),
        }

        for t in self.config.dtags:
            tag_name = t.name
            tag_type = t.type
            tag_size = t.size

            assert tag_type in typenames, "Invalid tag type; must be one of %r" % list(typenames)
            tag_class, tag_default, f = typenames[tag_type]
            tag_value = f(t.value)

            tag_address = t.addr
            logger.debug("tag address: %s", tag_address)

            path, attribute = None, None
            if tag_address:
                # Resolve the @cls/ins/att, and optionally [elm] or /elm
                segments, elm, cnt = device.parse_path_elements('@' + tag_address)
                assert not cnt or cnt == 1, \
                    "A Tag may be specified to indicate a single element: %s" % (tag_address)
                path = {'segment': segments}
                cls, ins, att = device.resolve(path, attribute=True)
                assert ins > 0, "Cannot specify the Class' instance for a tag's address"
                elm = device.resolve_element(path)
                # Look thru defined tags for one assigned to same cls/ins/att (maybe different elm);
                # must be same type/size.
                for tn, te in dict.items(self.tags):
                    if not te['path']:
                        continue  # Ignore tags w/o pre-defined path...
                    if device.resolve(te['path'], attribute=True) == (cls, ins, att):
                        assert te.attribute.parser.__class__ is tag_class and len(te.attribute) == tag_size, \
                            "Incompatible Attribute types for tags %r and %r" % (tn, tag_name)
                        attribute = te.attribute
                        break

            if not attribute:
                # No Attribute found
                attribute = device.Attribute(tag_name, tag_class,
                                             default = (tag_value if tag_size == 1 else [tag_value] * tag_size))

            # Ready to create the tag and its Attribute (and error code to return, if any).  If tag_size
            # is 1, it will be a scalar Attribute.  Since the tag_name may contain '.', we don't want
            # the normal dotdict.__setitem__ resolution to parse it; use plain dict.__setitem__.
            logger.debug("Creating tag: %-14s%-10s %10s[%4d]", tag_name, '@' + tag_address if tag_address else '',
                       attribute.parser.__class__.__name__, len(attribute))
            tag_entry = cpppo.dotdict()
            tag_entry.attribute = attribute  # The Attribute (may be shared by multiple tags)
            tag_entry.path = path  # Desired Attribute path (may include element), or None
            tag_entry.error = 0x00
            dict.__setitem__(self.tags, tag_name, tag_entry)

    def start(self, host, port):
        srv_ctl = cpppo.dotdict()
        srv_ctl.control = cpppo.apidict(timeout=self.config.timeout)
        srv_ctl.control['done'] = False
        srv_ctl.control['disable'] = False
        srv_ctl.control.setdefault('latency', self.config.latency)

        options = cpppo.dotdict()
        options.setdefault('enip_process', logix.process)
        kwargs = dict(options, tags=self.tags, server=srv_ctl)

        tcp_mode = True if self.config.mode == 'tcp' else False
        udp_mode = True if self.config.mode == 'udp' else False

        logger.debug('ENIP server started on: %s:%d, mode: %s' % (host, port, self.config.mode))
        while not self.stopped:
            network.server_main(address=(host, port), target=self.handle,
                                kwargs=kwargs, idle_service=None,
                                udp=udp_mode, tcp=tcp_mode,
                                thread_factory=network.server_thread)

    def stop(self):
        logger.debug('Stopping ENIP server')
        self.stopped = True


if __name__ == '__main__':
    import conpot, os
    dir_name = os.path.dirname(conpot.__file__)
    template = dir_name + '/templates/default/enip/enip.xml'
    enip_server_implicit = EnipServer(template, None, None)
    enip_server_implicit.config.mode = 'udp'
    enip_server_implicit.port = 60002
    try:
        enip_server_implicit.start(enip_server_implicit.addr, enip_server_implicit.port)
    except KeyboardInterrupt:
        enip_server_implicit.stop()



# Copyright (C) 2015  Peter Sooky <xsooky00@stud.fit.vutbr.cz>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent
from gevent import monkey; gevent.monkey.patch_all()

import unittest
from gevent import socket
import os
import conpot
import conpot.core as conpot_core
from conpot.protocols.bacnet import bacnet_server
from collections import namedtuple
from bacpypes.pdu import GlobalBroadcast, PDU
from bacpypes.apdu import APDU, WhoIsRequest, IAmRequest, IHaveRequest,  WhoHasObject,  WhoHasRequest, \
    ReadPropertyRequest, ReadPropertyACK, AtomicReadFileRequest, AuthenticateRequest
from bacpypes.constructeddata import Any
from bacpypes.primitivedata import Real


class TestBACnetServer(unittest.TestCase):

    """
        All tests are executed in a similar way. We initiate a service request to the BACnet server and wait for response.
        Instead of decoding the response, we create an expected response. We encode the expected response and compare the
        two encoded data.
    """

    def setUp(self):
        # clean up before we start...
        conpot_core.get_sessionManager().purge_sessions()

        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)
        args = namedtuple('FakeArgs', '')
        self.bacnet_server = bacnet_server.BacnetServer(self.dir_name + '/templates/default/bacnet/bacnet.xml',
                                                        'none', args)
        self.server_greenlet = gevent.spawn(self.bacnet_server.start, '127.0.0.1', 0)
        gevent.sleep(1)
        # initialize the databus
        self.databus = conpot_core.get_databus()
        self.databus.initialize(self.dir_name + '/templates/default/template.xml')

    def tearDown(self):
        self.bacnet_server.stop()
        gevent.joinall([self.server_greenlet])
        # tidy up (again)...
        conpot_core.get_sessionManager().purge_sessions()

    def test_whoIs(self):
        request = WhoIsRequest(deviceInstanceRangeLowLimit=500, deviceInstanceRangeHighLimit=50000)
        apdu = APDU()
        request.encode(apdu)
        pdu = PDU()
        apdu.encode(pdu)
        buf_size = 1024
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.sendto(pdu.pduData, ('127.0.0.1', self.bacnet_server.server.server_port))
        data = s.recvfrom(buf_size)
        s.close()
        received_data = data[0]

        expected = IAmRequest()
        expected.pduDestination = GlobalBroadcast()
        expected.iAmDeviceIdentifier = 36113
        expected.maxAPDULengthAccepted = 1024
        expected.segmentationSupported = 'segmentedBoth'
        expected.vendorID = 15

        exp_apdu = APDU()
        expected.encode(exp_apdu)
        exp_pdu = PDU()
        exp_apdu.encode(exp_pdu)

        self.assertEqual(exp_pdu.pduData, received_data)

    def test_whoHas(self):
        request_object = WhoHasObject()
        request_object.objectIdentifier = ('binaryInput', 12)
        request = WhoHasRequest(object=request_object)
        apdu = APDU()
        request.encode(apdu)
        pdu = PDU()
        apdu.encode(pdu)
        buf_size = 1024
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.sendto(pdu.pduData, ('127.0.0.1', self.bacnet_server.server.server_port))
        data = s.recvfrom(buf_size)
        s.close()
        received_data = data[0]

        expected = IHaveRequest()
        expected.pduDestination = GlobalBroadcast()
        expected.deviceIdentifier = 36113
        expected.objectIdentifier = 12
        expected.objectName = 'BI 01'

        exp_apdu = APDU()
        expected.encode(exp_apdu)
        exp_pdu = PDU()
        exp_apdu.encode(exp_pdu)
        self.assertEqual(exp_pdu.pduData, received_data)

    def test_readProperty(self):

        request = ReadPropertyRequest(objectIdentifier=('analogInput', 14), propertyIdentifier=85)
        request.apduMaxResp = 1024
        request.apduInvokeID = 101
        apdu = APDU()
        request.encode(apdu)
        pdu = PDU()
        apdu.encode(pdu)
        buf_size = 1024
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.sendto(pdu.pduData, ('127.0.0.1', self.bacnet_server.server.server_port))
        data = s.recvfrom(buf_size)
        s.close()
        received_data = data[0]

        expected = ReadPropertyACK()
        expected.pduDestination = GlobalBroadcast()
        expected.apduInvokeID = 101
        expected.objectIdentifier = 14
        expected.objectName = 'AI 01'
        expected.propertyIdentifier = 85
        expected.propertyValue = Any(Real(68.0))

        exp_apdu = APDU()
        expected.encode(exp_apdu)
        exp_pdu = PDU()
        exp_apdu.encode(exp_pdu)

        self.assertEqual(exp_pdu.pduData, received_data)

    def test_no_response_requests(self):
        """When the request has apduType not 0x01, no reply should be returned from Conpot"""
        request = ReadPropertyRequest(objectIdentifier=('analogInput', 14), propertyIdentifier=85)
        request.pduData = bytearray(b'test_data')
        request.apduMaxResp = 1024
        request.apduInvokeID = 101
        # Build requests - Confirmed, simple ack pdu, complex ack pdu, error pdu - etc.
        test_requests = list()

        for i in range(2, 8):
            if i not in {1, 3, 4}:
                request.apduType = i
                if i == 2:
                    # when apdu.apduType is 2 - we have SimpleAckPDU
                    # set the apduInvokeID and apduService
                    request.apduService = 8
                elif i == 5:
                    # when apdu.apduType is 5 - we have ErrorPDU
                    # set the apduInvokeID and apduService
                    request.apduService = 8
                elif i == 6:
                    # when apdu.apduType is 6 - we have RejectPDU
                    # set the apduInvokeID and apduAbortRejectReason
                    request.apduAbortRejectReason = 9
                else:
                    # when apdu.apduType is 7 - we have AbortPDU
                    # set the apduInvokeID and apduAbortRejectReason
                    request.apduAbortRejectReason = 9

                test_requests.append(request)
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        buf_size = 1024
        [s.sendto(i.pduData, ('127.0.0.1', self.bacnet_server.server.server_port)) for i in test_requests]
        results = None
        with gevent.Timeout(1, False):
            results = [s.recvfrom(buf_size) for i in range(len(test_requests))]
        self.assertIsNone(results)


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2013  Daniel creo Haslinger <creo-conpot@blackmesa.at>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import time
import random

from datetime import datetime

from html.parser import HTMLParser
from socketserver import ThreadingMixIn

import http.server
import http.client
import os
from lxml import etree
from conpot.helpers import str_to_bytes
import conpot.core as conpot_core
import gevent


logger = logging.getLogger(__name__)


class HTTPServer(http.server.BaseHTTPRequestHandler):

    def log(self, version, request_type, addr, request, response=None):

        session = conpot_core.get_session('http', addr[0], addr[1])

        log_dict = {'remote': addr,
                    'timestamp': datetime.utcnow(),
                    'data_type': 'http',
                    'dst_port': self.server.server_port,
                    'data': {0: {'request': '{0} {1}: {2}'.format(version, request_type, request)}}}

        logger.info('%s %s request from %s: %s. %s', version, request_type, addr, request, session.id)

        if response:
            logger.info('%s response to %s: %s. %s', version, addr, response, session.id)
            log_dict['data'][0]['response'] = '{0} response: {1}'.format(version, response)
            session.add_event({'request': str(request), 'response': str(response)})
        else:
            session.add_event({'request': str(request)})

        # FIXME: Proper logging

    def get_entity_headers(self, rqfilename, headers, configuration):

        xml_headers = configuration.xpath(
            '//http/htdocs/node[@name="' + rqfilename + '"]/headers/*'
        )

        if xml_headers:

            # retrieve all headers assigned to this entity
            for header in xml_headers:
                headers.append((header.attrib['name'], header.text))

        return headers

    def get_trigger_appendix(self, rqfilename, rqparams, configuration):

        xml_triggers = configuration.xpath(
            '//http/htdocs/node[@name="' + rqfilename + '"]/triggers/*'
        )

        if xml_triggers:
            paramlist = rqparams.split('&')

            # retrieve all subselect triggers assigned to this entity
            for triggers in xml_triggers:

                triggerlist = triggers.text.split(';')
                trigger_missed = False

                for trigger in triggerlist:
                    if not trigger in paramlist:
                        trigger_missed = True

                if not trigger_missed:
                    return triggers.attrib['appendix']

        return None

    def get_entity_trailers(self, rqfilename, configuration):

        trailers = []
        xml_trailers = configuration.xpath(
            '//http/htdocs/node[@name="' + rqfilename + '"]/trailers/*'
        )

        if xml_trailers:

            # retrieve all headers assigned to this entity
            for trailer in xml_trailers:
                trailers.append((trailer.attrib['name'], trailer.text))

        return trailers

    def get_status_headers(self, status, headers, configuration):

        xml_headers = configuration.xpath('//http/statuscodes/status[@name="' +
                                          str(status) + '"]/headers/*')

        if xml_headers:

            # retrieve all headers assigned to this status
            for header in xml_headers:
                headers.append((header.attrib['name'], header.text))

        return headers

    def get_status_trailers(self, status, configuration):

        trailers = []
        xml_trailers = configuration.xpath(
            '//http/statuscodes/status[@name="' + str(status) + '"]/trailers/*'
        )

        if xml_trailers:

            # retrieve all trailers assigned to this status
            for trailer in xml_trailers:
                trailers.append((trailer.attrib['name'], trailer.text))

        return trailers

    def send_response(self, code, message=None):
        """Send the response header and log the response code.
        This function is overloaded to change the behaviour when
        loggers and sending default headers.
        """

        # replace integrated loggers with conpot logger..
        # self.log_request(code)

        if message is None:
            if code in self.responses:
                message = self.responses[code][0]
            else:
                message = ''

        if self.request_version != 'HTTP/0.9':
            msg = str_to_bytes("{} {} {}\r\n".format(self.protocol_version, code, message))
            self.wfile.write(msg)

        # the following two headers are omitted, which is why we override
        # send_response() at all. We do this one on our own...

        # - self.send_header('Server', self.version_string())
        # - self.send_header('Date', self.date_time_string())

    def substitute_template_fields(self, payload):

        # initialize parser with our payload
        parser = TemplateParser(payload)

        # triggers the parser, just in case of open / incomplete tags..
        parser.close()

        # retrieve and return (substituted) payload
        return parser.payload

    def load_status(self, status, requeststring, requestheaders, headers, configuration, docpath, method='GET', body=None):
        """Retrieves headers and payload for a given status code.
           Certain status codes can be configured to forward the
           request to a remote system. If not available, generate
           a minimal response"""

        # handle PROXY tag
        entity_proxy = configuration.xpath('//http/statuscodes/status[@name="' +
                                           str(status) +
                                           '"]/proxy')

        if entity_proxy:
            source = 'proxy'
            target = entity_proxy[0].xpath('./text()')[0]
        else:
            source = 'filesystem'

        # handle TARPIT tag
        entity_tarpit = configuration.xpath(
            '//http/statuscodes/status[@name="' + str(status) + '"]/tarpit'
        )

        if entity_tarpit:
            tarpit = self.server.config_sanitize_tarpit(entity_tarpit[0].xpath('./text()')[0])
        else:
            tarpit = None

        # check if we have to delay further actions due to global or local TARPIT configuration
        if tarpit is not None:
            # this node has its own delay configuration
            self.server.do_tarpit(tarpit)
        else:
            # no delay configuration for this node. check for global latency
            if self.server.tarpit is not None:
                # fall back to the globally configured latency
                self.server.do_tarpit(self.server.tarpit)

        # If the requested resource resides on our filesystem,
        # we try retrieve all metadata and the resource itself from there.
        if source == 'filesystem':

            # retrieve headers from entities configuration block
            headers = self.get_status_headers(status, headers, configuration)

            # retrieve headers from entities configuration block
            trailers = self.get_status_trailers(status, configuration)

            # retrieve payload directly from filesystem, if possible.
            # If this is not possible, return an empty, zero sized string.
            try:
                if not isinstance(status, int):
                    status = status.value
                with open(os.path.join(docpath, 'statuscodes', str(int(status)) + '.status'), 'rb') as f:
                    payload = f.read()

            except IOError as e:
                logger.exception('%s', e)
                payload = ''

            # there might be template data that can be substituted within the
            # payload. We only substitute data that is going to be displayed
            # by the browser:

            # perform template substitution on payload
            payload = self.substitute_template_fields(payload)

            # How do we transport the content?
            chunked_transfer = configuration.xpath('//http/htdocs/node[@name="' +
                                                   str(status) + '"]/chunks')

            if chunked_transfer:
                # Append a chunked transfer encoding header
                headers.append(('Transfer-Encoding', 'chunked'))
                chunks = str(chunked_transfer[0].xpath('./text()')[0])
            else:
                # Calculate and append a content length header
                headers.append(('Content-Length', payload.__len__()))
                chunks = '0'

            return status, headers, trailers, payload, chunks

        # the requested status code is configured to forward the
        # originally targeted resource to a remote system.

        elif source == 'proxy':

            # open a connection to the remote system.
            # If something goes wrong, fall back to 503.

            # NOTE: we use try:except here because there is no perfect
            # platform independent way to check file accessibility.

            trailers = []
            chunks = '0'

            try:
                # Modify a few headers to fit our new destination and the fact
                # that we're proxying while being unaware of any session foo..
                requestheaders['Host'] = target
                requestheaders['Connection'] = 'close'

                remotestatus = 0
                conn = http.client.HTTPConnection(target)
                conn.request(method, requeststring, body, dict(requestheaders))
                response = conn.getresponse()

                remotestatus = int(response.status)
                headers = response.getheaders()   # We REPLACE the headers to avoid duplicates!
                payload = response.read()

                # WORKAROUND: to get around a strange httplib-behaviour when it comes
                # to chunked transfer encoding, we replace the chunked-header with a
                # valid Content-Length header:

                for i, header in enumerate(headers):

                    if header[0].lower() == 'transfer-encoding' and header[1].lower() == 'chunked':
                        del headers[i]
                        break

                status = remotestatus

            except:

                # before falling back to 503, we check if we are ALREADY dealing with a 503
                # to prevent an infinite request handling loop...

                if status != 503:

                    # we're handling another error here.
                    # generate a 503 response from configuration.
                    (status, headers, trailers, payload, chunks) = self.load_status(503,
                                                                                    requeststring,
                                                                                    self.headers,
                                                                                    headers,
                                                                                    configuration,
                                                                                    docpath)

                else:

                    # oops, we're heading towards an infinite loop here,
                    # generate a minimal 503 response regardless of the configuration.
                    status = 503
                    payload = ''
                    chunks = '0'
                    headers.append(('Content-Length', 0))

            return status, headers, trailers, payload, chunks

    def load_entity(self, requeststring, headers, configuration, docpath):
        """
        Retrieves status, headers and payload for a given entity, that
        can be stored either local or on a remote system
        """

        # extract filename and GET parameters from request string
        rqfilename = requeststring.partition('?')[0]
        rqparams = requeststring.partition('?')[2]

        # handle ALIAS tag
        entity_alias = configuration.xpath(
            '//http/htdocs/node[@name="' + rqfilename + '"]/alias'
        )
        if entity_alias:
            rqfilename = entity_alias[0].xpath('./text()')[0]

        # handle SUBSELECT tag
        rqfilename_appendix = self.get_trigger_appendix(rqfilename, rqparams, configuration)
        if rqfilename_appendix:
            rqfilename += '_' + rqfilename_appendix

        # handle PROXY tag
        entity_proxy = configuration.xpath(
            '//http/htdocs/node[@name="' + rqfilename + '"]/proxy'
        )
        if entity_proxy:
            source = 'proxy'
            target = entity_proxy[0].xpath('./text()')[0]
        else:
            source = 'filesystem'

        # handle TARPIT tag
        entity_tarpit = configuration.xpath(
            '//http/htdocs/node[@name="' + rqfilename + '"]/tarpit'
        )
        if entity_tarpit:
            tarpit = self.server.config_sanitize_tarpit(entity_tarpit[0].xpath('./text()')[0])
        else:
            tarpit = None

        # check if we have to delay further actions due to global or local TARPIT configuration
        if tarpit is not None:
            # this node has its own delay configuration
            self.server.do_tarpit(tarpit)
        else:
            # no delay configuration for this node. check for global latency
            if self.server.tarpit is not None:
                # fall back to the globally configured latency
                self.server.do_tarpit(self.server.tarpit)

        # If the requested resource resides on our filesystem,
        # we try retrieve all metadata and the resource itself from there.
        if source == 'filesystem':

            # handle STATUS tag
            # ( filesystem only, since proxied requests come with their own status )
            entity_status = configuration.xpath(
                '//http/htdocs/node[@name="' + rqfilename + '"]/status'
            )
            if entity_status:
                status = int(entity_status[0].xpath('./text()')[0])
            else:
                status = 200

            # retrieve headers from entities configuration block
            headers = self.get_entity_headers(rqfilename, headers, configuration)

            # retrieve trailers from entities configuration block
            trailers = self.get_entity_trailers(rqfilename, configuration)

            # retrieve payload directly from filesystem, if possible.
            # If this is not possible, return an empty, zero sized string.
            if os.path.isabs(rqfilename):
                relrqfilename = rqfilename[1:]
            else:
                relrqfilename = rqfilename

            try:
                with open(os.path.join(docpath, 'htdocs', relrqfilename), 'rb') as f:
                    payload = f.read()

            except IOError as e:
                if not os.path.isdir(os.path.join(docpath, 'htdocs', relrqfilename)):
                    logger.error('Failed to get template content: %s', e)
                payload = ''

            # there might be template data that can be substituted within the
            # payload. We only substitute data that is going to be displayed
            # by the browser:

            templated = False
            for header in headers:
                if header[0].lower() == 'content-type' and header[1].lower() == 'text/html':
                    templated = True

            if templated:
                # perform template substitution on payload
                payload = self.substitute_template_fields(payload)

            # How do we transport the content?
            chunked_transfer = configuration.xpath(
                '//http/htdocs/node[@name="' + rqfilename + '"]/chunks'
            )

            if chunked_transfer:
                # Calculate and append a chunked transfer encoding header
                headers.append(('Transfer-Encoding', 'chunked'))
                chunks = str(chunked_transfer[0].xpath('./text()')[0])
            else:
                # Calculate and append a content length header
                headers.append(('Content-Length', payload.__len__()))
                chunks = '0'

            return status, headers, trailers, payload, chunks

        # the requested resource resides on another server,
        # so we act as a proxy between client and target system

        elif source == 'proxy':

            # open a connection to the remote system.
            # If something goes wrong, fall back to 503

            trailers = []

            try:
                conn = http.client.HTTPConnection(target)
                conn.request("GET", requeststring)
                response = conn.getresponse()

                status = int(response.status)
                headers = response.getheaders()    # We REPLACE the headers to avoid duplicates!
                payload = response.read()
                chunks = '0'

            except:
                status = 503
                (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                                requeststring,
                                                                                self.headers,
                                                                                headers,
                                                                                configuration,
                                                                                docpath)

            return status, headers, trailers, payload, chunks

    def send_chunked(self, chunks, payload, trailers):
        """Send payload via chunked transfer encoding to the
        client, followed by eventual trailers."""

        chunk_list = chunks.split(',')
        pointer = 0
        for cwidth in chunk_list:
            cwidth = int(cwidth)
            # send chunk length indicator
            self.wfile.write(format(cwidth, 'x').upper() + "\r\n")
            # send chunk payload
            self.wfile.write(payload[pointer:pointer + cwidth] + "\r\n")
            pointer += cwidth

        # is there another chunk that has not been configured? Send it anyway for the sake of completeness..
        if len(payload) > pointer:
            # send chunk length indicator
            self.wfile.write(format(len(payload) - pointer, 'x').upper() + "\r\n")
            # send chunk payload
            self.wfile.write(payload[pointer:] + "\r\n")

        # we're done with the payload. Send a zero chunk as EOF indicator
        self.wfile.write('0'+"\r\n")

        # if there are trailing headers :-) we send them now..
        for trailer in trailers:
            self.wfile.write("%s: %s\r\n" % (trailer[0], trailer[1]))

        # and finally, the closing ceremony...
        self.wfile.write("\r\n")

    def send_error(self, code, message=None):
        """Send and log an error reply.
        This method is overloaded to make use of load_status()
        to allow handling of "Unsupported Method" errors.
        """

        headers = []
        headers.extend(self.server.global_headers)
        configuration = self.server.configuration
        docpath = self.server.docpath

        if not hasattr(self, 'headers'):
            self.headers = self.MessageClass(self.rfile, 0)

        trace_data_length = self.headers.get('content-length')
        unsupported_request_data = None

        if trace_data_length:
            unsupported_request_data = self.rfile.read(int(trace_data_length))

        # there are certain situations where variables are (not yet) registered
        # ( e.g. corrupted request syntax ). In this case, we set them manually.
        if hasattr(self, 'path') and self.path is not None:
            requeststring = self.path
        else:
            requeststring = ''
            self.path = None
            if message is not None:
                logger.info(message)

        # generate the appropriate status code, header and payload
        (status, headers, trailers, payload, chunks) = self.load_status(code,
                                                                        requeststring.partition('?')[0],
                                                                        self.headers,
                                                                        headers,
                                                                        configuration,
                                                                        docpath)

        # send http status to client
        self.send_response(status)

        # send all headers to client
        for header in headers:
            self.send_header(header[0], header[1])

        self.end_headers()

        # decide upon sending content as a whole or chunked
        if chunks == '0':
            # send payload as a whole to the client
            if type(payload) != bytes:
                payload = payload.encode()
            self.wfile.write(payload)
        else:
            # send payload in chunks to the client
            self.send_chunked(chunks, payload, trailers)

        # loggers
        self.log(self.request_version, self.command, self.client_address, (self.path,
                                                                           self.headers._headers,
                                                                           unsupported_request_data), status)

    def do_TRACE(self):
        """Handle TRACE requests."""

        # fetch configuration dependent variables from server instance
        headers = []
        headers.extend(self.server.global_headers)
        configuration = self.server.configuration
        docpath = self.server.docpath

        # retrieve TRACE body data
        # ( sticking to the HTTP protocol, there should not be any body in TRACE requests,
        #   an attacker could though use the body to inject data if not flushed correctly,
        #   which is done by accessing the data like we do now - just to be secure.. )

        trace_data_length = self.headers.get('content-length')
        trace_data = None

        if trace_data_length:
            trace_data = self.rfile.read(int(trace_data_length))

        # check configuration: are we allowed to use this method?
        if self.server.disable_method_trace is True:

            # Method disabled by configuration. Fall back to 501.
            status = 501
            (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                            self.path,
                                                                            self.headers,
                                                                            headers,
                                                                            configuration,
                                                                            docpath)

        else:

            # Method is enabled
            status = 200
            payload = ''
            headers.append(('Content-Type', 'message/http'))

            # Gather all request data and return it to sender..
            for rqheader in self.headers:
                payload = payload + str(rqheader) + ': ' + self.headers.get(rqheader) + "\n"

        # send initial HTTP status line to client
        self.send_response(status)

        # send all headers to client
        for header in headers:
            self.send_header(header[0], header[1])

        self.end_headers()

        # send payload (the actual content) to client
        if type(payload) != bytes:
            payload = payload.encode()
        self.wfile.write(payload)

        # loggers
        self.log(self.request_version,
                 self.command,
                 self.client_address,
                 (self.path, self.headers._headers, trace_data),
                 status)

    def do_HEAD(self):
        """Handle HEAD requests."""

        # fetch configuration dependent variables from server instance
        headers = list()
        headers.extend(self.server.global_headers)
        configuration = self.server.configuration
        docpath = self.server.docpath

        # retrieve HEAD body data
        # ( sticking to the HTTP protocol, there should not be any body in HEAD requests,
        #   an attacker could though use the body to inject data if not flushed correctly,
        #   which is done by accessing the data like we do now - just to be secure.. )

        head_data_length = self.headers.get('content-length')
        head_data = None

        if head_data_length:
            head_data = self.rfile.read(int(head_data_length))

        # check configuration: are we allowed to use this method?
        if self.server.disable_method_head is True:

            # Method disabled by configuration. Fall back to 501.
            status = 501
            (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                            self.path,
                                                                            self.headers,
                                                                            headers,
                                                                            configuration,
                                                                            docpath)

        else:

            # try to find a configuration item for this GET request
            entity_xml = configuration.xpath(
                '//http/htdocs/node[@name="'
                + self.path.partition('?')[0] + '"]'
            )

            if entity_xml:
                # A config item exists for this entity. Handle it..
                (status, headers, trailers, payload, chunks) = self.load_entity(self.path,
                                                                                headers,
                                                                                configuration,
                                                                                docpath)

            else:
                # No config item could be found. Fall back to a standard 404..
                status = 404
                (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                                self.path,
                                                                                self.headers,
                                                                                headers,
                                                                                configuration,
                                                                                docpath)

        # send initial HTTP status line to client
        self.send_response(status)

        # send all headers to client
        for header in headers:
            self.send_header(header[0], header[1])

        self.end_headers()

        # loggers
        self.log(self.request_version,
                 self.command,
                 self.client_address,
                 (self.path, self.headers._headers, head_data),
                 status)

    def do_OPTIONS(self):
        """Handle OPTIONS requests."""

        # fetch configuration dependent variables from server instance
        headers = []
        headers.extend(self.server.global_headers)
        configuration = self.server.configuration
        docpath = self.server.docpath

        # retrieve OPTIONS body data
        # ( sticking to the HTTP protocol, there should not be any body in HEAD requests,
        #   an attacker could though use the body to inject data if not flushed correctly,
        #   which is done by accessing the data like we do now - just to be secure.. )

        options_data_length = self.headers.get('content-length')
        options_data = None

        if options_data_length:
            options_data = self.rfile.read(int(options_data_length))

        # check configuration: are we allowed to use this method?
        if self.server.disable_method_options is True:

            # Method disabled by configuration. Fall back to 501.
            status = 501
            (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                            self.path,
                                                                            self.headers,
                                                                            headers,
                                                                            configuration,
                                                                            docpath)

        else:

            status = 200
            payload = ''

            # Add ALLOW header to response. GET, POST and OPTIONS are static, HEAD and TRACE are dynamic
            allowed_methods = 'GET'

            if self.server.disable_method_head is False:
                # add head to list of allowed methods
                allowed_methods += ',HEAD'

            allowed_methods += ',POST,OPTIONS'

            if self.server.disable_method_trace is False:
                allowed_methods += ',TRACE'

            headers.append(('Allow', allowed_methods))

            # Calculate and append a content length header
            headers.append(('Content-Length', payload.__len__()))

            # Append CC header
            headers.append(('Connection', 'close'))

            # Append CT header
            headers.append(('Content-Type', 'text/html'))

        # send initial HTTP status line to client
        self.send_response(status)

        # send all headers to client
        for header in headers:
            self.send_header(header[0], header[1])

        self.end_headers()

        # loggers
        self.log(self.request_version,
                 self.command,
                 self.client_address,
                 (self.path, self.headers._headers, options_data),
                 status)

    def do_GET(self):
        """Handle GET requests"""

        # fetch configuration dependent variables from server instance
        headers = []
        headers.extend(self.server.global_headers)
        configuration = self.server.configuration
        docpath = self.server.docpath

        # retrieve GET body data
        # ( sticking to the HTTP protocol, there should not be any body in GET requests,
        #   an attacker could though use the body to inject data if not flushed correctly,
        #   which is done by accessing the data like we do now - just to be secure.. )

        get_data_length = self.headers.get('content-length')
        get_data = None

        if get_data_length:
            get_data = self.rfile.read(int(get_data_length))

        # try to find a configuration item for this GET request
        entity_xml = configuration.xpath(
            '//http/htdocs/node[@name="' + self.path.partition('?')[0] + '"]'
        )

        if entity_xml:
            # A config item exists for this entity. Handle it..
            (status, headers, trailers, payload, chunks) = self.load_entity(self.path,
                                                                            headers,
                                                                            configuration,
                                                                            docpath)

        else:
            # No config item could be found. Fall back to a standard 404..
            status = 404
            (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                            self.path,
                                                                            self.headers,
                                                                            headers,
                                                                            configuration,
                                                                            docpath,
                                                                            'GET')

        # send initial HTTP status line to client
        self.send_response(status)

        # send all headers to client
        for header in headers:
            self.send_header(header[0], header[1])

        self.end_headers()

        # decide upon sending content as a whole or chunked
        if chunks == '0':
            # send payload as a whole to the client
            self.wfile.write(str_to_bytes(payload))
        else:
            # send payload in chunks to the client
            self.send_chunked(chunks, payload, trailers)

        # loggers
        self.log(self.request_version,
                 self.command,
                 self.client_address,
                 (self.path, self.headers._headers, get_data),
                 status)

    def do_POST(self):
        """Handle POST requests"""

        # fetch configuration dependent variables from server instance
        headers = list()
        headers.extend(self.server.global_headers)
        configuration = self.server.configuration
        docpath = self.server.docpath

        # retrieve POST data ( important to flush request buffers )
        post_data_length = self.headers.get('content-length')
        post_data = None

        if post_data_length:
            post_data = self.rfile.read(int(post_data_length))

        # try to find a configuration item for this POST request
        entity_xml = configuration.xpath(
            '//http/htdocs/node[@name="' + self.path.partition('?')[0] + '"]'
        )

        if entity_xml:
            # A config item exists for this entity. Handle it..
            (status, headers, trailers, payload, chunks) = self.load_entity(self.path,
                                                                            headers,
                                                                            configuration,
                                                                            docpath)

        else:
            # No config item could be found. Fall back to a standard 404..
            status = 404
            (status, headers, trailers, payload, chunks) = self.load_status(status,
                                                                            self.path,
                                                                            self.headers,
                                                                            headers,
                                                                            configuration,
                                                                            docpath,
                                                                            'POST',
                                                                            post_data)

        # send initial HTTP status line to client
        self.send_response(status)

        # send all headers to client
        for header in headers:
            self.send_header(header[0], header[1])

        self.end_headers()

        # decide upon sending content as a whole or chunked
        if chunks == '0':
            # send payload as a whole to the client
            if type(payload) != bytes:
                payload = payload.encode()
            self.wfile.write(payload)
        else:
            # send payload in chunks to the client
            self.send_chunked(chunks, payload, trailers)

        # loggers
        self.log(self.request_version,
                 self.command,
                 self.client_address,
                 (self.path, self.headers._headers, post_data),
                 status)


class TemplateParser(HTMLParser):
    def __init__(self, data):
        self.databus = conpot_core.get_databus()
        if type(data) == bytes:
            data = data.decode()
        self.data = data
        HTMLParser.__init__(self)
        self.payload = self.data
        self.feed(self.data)

    def handle_startendtag(self, tag, attrs):
        """ handles template tags provided in XHTML notation.

            Expected format:    <condata source="(engine)" key="(descriptor)" />
            Example:            <condata source="databus" key="SystemDescription" />

            at the moment, the parser is space- and case-sensitive(!),
            this could be improved by using REGEX for replacing the template tags
            with actual values.
        """

        source = ''
        key = ''

        # only parse tags that are conpot template tags ( <condata /> )
        if tag == 'condata':

            # initialize original tag (needed for value replacement)
            origin = '<' + tag

            for attribute in attrs:

                # extend original tag
                origin = origin + ' ' + attribute[0] + '="' + attribute[1] + '"'

                # fill variables with all meta information needed to
                # gather actual data from the other engines (databus, modbus, ..)
                if attribute[0] == 'source':
                    source = attribute[1]
                elif attribute[0] == 'key':
                    key = attribute[1]

            # finalize original tag
            origin += ' />'

            # we really need a key in order to do our work..
            if key:
                # deal with databus powered tags:
                if source == 'databus':
                    self.result = self.databus.get_value(key)
                    self.payload = self.payload.replace(origin, str(self.result))

                # deal with eval powered tags:
                elif source == 'eval':
                    result = ''
                    # evaluate key
                    try:
                        result = eval(key)
                    except Exception as e:
                        logger.exception(e)
                    self.payload = self.payload.replace(origin, result)


class ThreadedHTTPServer(ThreadingMixIn, http.server.HTTPServer):
    """Handle requests in a separate thread."""


class SubHTTPServer(ThreadedHTTPServer):
    """this class is necessary to allow passing custom request handler into
       the RequestHandlerClass"""
    daemon_threads = True

    def __init__(self, server_address, RequestHandlerClass, template, docpath):
        http.server.HTTPServer.__init__(self, server_address, RequestHandlerClass)

        self.docpath = docpath

        # default configuration
        self.update_header_date = True             # this preserves authenticity
        self.disable_method_head = False
        self.disable_method_trace = False
        self.disable_method_options = False
        self.tarpit = '0'

        # load the configuration from template and parse it
        # for the first time in order to reduce further handling..
        self.configuration = etree.parse(template)

        xml_config = self.configuration.xpath('//http/global/config/*')
        if xml_config:

            # retrieve all global configuration entities
            for entity in xml_config:

                if entity.attrib['name'] == 'protocol_version':
                    RequestHandlerClass.protocol_version = entity.text

                elif entity.attrib['name'] == 'update_header_date':
                    if entity.text.lower() == 'false':
                        # DATE header auto update disabled by configuration
                        self.update_header_date = False
                    elif entity.text.lower() == 'true':
                        # DATE header auto update enabled by configuration
                        self.update_header_date = True

                elif entity.attrib['name'] == 'disable_method_head':
                    if entity.text.lower() == 'false':
                        # HEAD method enabled by configuration
                        self.disable_method_head = False
                    elif entity.text.lower() == 'true':
                        # HEAD method disabled by configuration
                        self.disable_method_head = True

                elif entity.attrib['name'] == 'disable_method_trace':
                    if entity.text.lower() == 'false':
                        # TRACE method enabled by configuration
                        self.disable_method_trace = False
                    elif entity.text.lower() == 'true':
                        # TRACE method disabled by configuration
                        self.disable_method_trace = True

                elif entity.attrib['name'] == 'disable_method_options':
                    if entity.text.lower() == 'false':
                        # OPTIONS method enabled by configuration
                        self.disable_method_options = False
                    elif entity.text.lower() == 'true':
                        # OPTIONS method disabled by configuration
                        self.disable_method_options = True

                elif entity.attrib['name'] == 'tarpit':
                    if entity.text:
                        self.tarpit = self.config_sanitize_tarpit(entity.text)

        # load global headers from XML
        self.global_headers = []
        xml_headers = self.configuration.xpath('//http/global/headers/*')
        if xml_headers:

            # retrieve all headers assigned to this status code
            for header in xml_headers:
                if header.attrib['name'].lower() == 'date' and self.update_header_date is True:
                    # All HTTP date/time stamps MUST be represented in Greenwich Mean Time (GMT),
                    # without exception ( RFC-2616 )
                    self.global_headers.append((header.attrib['name'],
                                                time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime())))
                else:
                    self.global_headers.append((header.attrib['name'], header.text))

    def config_sanitize_tarpit(self, value):

        # checks tarpit value for being either a single int or float,
        # or a series of two concatenated integers and/or floats seperated by semicolon and returns
        # either the (sanitized) value or zero.

        if value is not None:

            x, _, y = value.partition(';')

            try:
                _ = float(x)
            except ValueError:
                # first value is invalid, ignore the whole setting.
                logger.error("Invalid tarpit value: '%s'. Assuming no latency.", value)
                return '0;0'

            try:
                _ = float(y)
                # both values are fine.
                return value
            except ValueError:
                # second value is invalid, use the first one.
                return x

        else:
            return '0;0'

    def do_tarpit(self, delay):

        # sleeps the thread for $delay ( should be either 1 float to apply a static period of time to sleep,
        # or 2 floats seperated by semicolon to sleep a randomized period of time determined by ( rand[x;y] )

        lbound, _, ubound = delay.partition(";")

        if not lbound or lbound is None:
            # no lower boundary found. Assume zero latency
            pass
        elif not ubound or ubound is None:
            # no upper boundary found. Assume static latency
            gevent.sleep(float(lbound))
        else:
            # both boundaries found. Assume random latency between lbound and ubound
            gevent.sleep(random.uniform(float(lbound), float(ubound)))


class CommandResponder(object):

    def __init__(self, host, port, template, docpath):

        # Create HTTP server class
        self.httpd = SubHTTPServer((host, port), HTTPServer, template, docpath)
        self.server_port = self.httpd.server_port

    def serve_forever(self):
        self.httpd.serve_forever()

    def stop(self):
        logging.info("HTTP server will shut down gracefully as soon as all connections are closed.")
        self.httpd.shutdown()


# References: S7_300-400_full_reference_handbook_ENGLISH.pdf
#             http://www.bj-ig.de/147.html
#             https://code.google.com/p/plcscan/source/browse/trunk/s7.py


from struct import *

import struct
import conpot.core as conpot_core
from conpot.helpers import str_to_bytes
from conpot.protocols.s7comm.exceptions import AssembleException, ParseException


# S7 packet
class S7(object):

    ssl_lists = {}

    def __init__(self, pdu_type=0, reserved=0, request_id=0, result_info=0, parameters='', data=''):
        self.magic = 0x32
        self.pdu_type = pdu_type
        self.reserved = reserved
        self.request_id = request_id
        # sometimes "parameters" happen to be of type int, not str
        self.param_length = len(str(parameters))
        self.data_length = len(data)
        self.result_info = result_info
        self.parameters = parameters
        self.data = data

        # param codes (http://www.bj-ig.de/147.html):
        # maps request types to methods
        self.param_mapping = {0x00: ('diagnostics', self.request_diagnostics),
                              0x04: ('read', self.request_not_implemented),
                              0x05: ('write', self.request_not_implemented),
                              0x1a: ('request_download', self.request_not_implemented),
                              0x1b: ('download_block', self.request_not_implemented),
                              0x1c: ('end_download', self.request_not_implemented),
                              0x1d: ('start_upload', self.request_not_implemented),
                              0x1e: ('upload', self.request_not_implemented),
                              0x1f: ('end_upload', self.request_not_implemented),
                              0x28: ('insert_block', self.request_not_implemented)}

        # maps valid pdu codes to name
        self.pdu_mapping = {0x01: set('request_pdu'),
                            0x02: set('known_but_unindentified_pdu'),
                            0x03: set('response_pdu'),
                            0x07: set('system_status_list')}

        self.data_bus = conpot_core.get_databus()

    def __len__(self):

        if self.pdu_type in (2, 3):
            return 12 + int(self.param_length) + int(self.data_length)
        else:
            return 10 + int(self.param_length) + int(self.data_length)

    def handle(self):
        if self.param in self.param_mapping:
            # direct execution to the correct method based on the param
            return self.param_mapping[self.param][1]()

    def request_not_implemented(self):
        raise ParseException('s7comm', 'request not implemented in honeypot yet.')

    def pack(self):

        if self.pdu_type not in self.pdu_mapping:
            raise AssembleException('s7comm', 'invalid or unsupported pdu type')
        elif self.pdu_type in (2, 3):
            # type 2 and 3 feature an additional RESULT INFORMATION header
            return pack('!BBHHHHH', self.magic, self.pdu_type, self.reserved, self.request_id, self.param_length,
                        self.data_length, self.result_info) + str_to_bytes(self.parameters) + str_to_bytes(self.data)
        else:
            return pack('!BBHHHH', self.magic, self.pdu_type, self.reserved, self.request_id, self.param_length,
                        self.data_length) + str_to_bytes(self.parameters) + str_to_bytes(self.data)

    def parse(self, packet):

        # dissect fixed header
        try:
            fixed_header = unpack('!BBHHHH', packet[:10])
        except struct.error:
            raise ParseException('s7comm', 'malformed fixed packet header structure')

        self.magic = int(fixed_header[0])

        if self.magic != 0x32:
            raise ParseException('s7comm', 'bad magic number, expected 0x32 but got {0}.'.format(self.magic))

        self.pdu_type = fixed_header[1]
        self.reserved = fixed_header[2]
        self.request_id = fixed_header[3]
        self.param_length = fixed_header[4]
        self.data_length = fixed_header[5]

        # dissect variable header

        if self.pdu_type in (2, 3):
            # type 2 and 3 feature an additional RESULT INFORMATION header
            self.result_info = unpack('!H', packet[10:12])
            header_offset = 2
        else:
            header_offset = 0

        self.parameters = packet[10 + header_offset:10 + header_offset + self.param_length]
        self.data = packet[10 + header_offset + self.param_length:10 + header_offset + self.param_length +
                           self.data_length]

        try:
            self.param = unpack('!B', self.parameters[:1])[0]
        except:
            raise ParseException('s7comm', 'invalid packet')

        return self

    # SSL/SZL System Status List/Systemzustandsliste
    def request_diagnostics(self):

        # semi-check
        try:
            unpack('!BBBBBBBB', self.parameters[:8])
        except struct.error:
            raise ParseException('s7comm', 'malformed SSL/SZL parameter structure')

        chunk = self.data
        chunk_id = 0

        while chunk:
            try:
                ssl_chunk_header = unpack('!BBH', chunk[:4])
            except struct.error:
                raise ParseException('s7comm', 'malformed SSL/SZL data structure')

            # dissect data blocks

            data_error_code = ssl_chunk_header[0]
            data_data_type = ssl_chunk_header[1]
            data_next_bytes = ssl_chunk_header[2]
            data_ssl_id = ''
            data_ssl_index = ''
            data_ssl_unknown = ''

            if data_next_bytes > 0:
                data_ssl_id = unpack('!H', chunk[4:6])[0]

            if data_next_bytes > 1:
                data_ssl_index = unpack('!H', chunk[6:8])[0]

            if data_next_bytes > 2:
                data_ssl_unknown = chunk[8:4 + data_next_bytes]

            # map request ssl to method
            if hasattr(self, 'request_ssl_{0}'.format(data_ssl_id)):
                m = getattr(self, 'request_ssl_{0}'.format(data_ssl_id))
                description, params, data = m(data_ssl_index)
                return params, data

            chunk = chunk[4 + data_next_bytes:]
            chunk_id += 1

        return 0x00, 0x00

    # W#16#xy11 - module identification
    def request_ssl_17(self, data_ssl_index):

        # just for convenience
        current_ssl = S7.ssl_lists['W#16#xy11']

        if data_ssl_index == 1:    # 0x0001 - component identification

            ssl_index_description = 'Component identification'

            ssl_resp_data = pack('!HHHHH20sHHH',
                                 17,  # 1  WORD   ( ID )
                                 data_ssl_index,  # 1  WORD   ( Index )
                                 28,  # 1  WORD   ( Length of payload after element count )
                                 0x01,  # 1  WORD   ( 1 element follows )
                                 data_ssl_index,  # 1  WORD   ( Data Index )
                                 str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0001'])),
                                 # 10 WORDS  ( MLFB of component: 20 bytes => 19 chars + 1 blank (0x20) )
                                 0x0,  # 1  WORD   ( RESERVED )
                                 0x0,  # 1  WORD   ( Output state of component )
                                 0x0)                       # 1  WORD   ( RESERVED )

            ssl_resp_head = pack('!BBH',
                                 0xff,  # 1  BYTE   ( Data Error Code. 0xFF = OK )
                                 0x09,  # 1  BYTE   ( Data Type. 0x09 = Char/String )
                                 len(ssl_resp_data))        # 1  WORD   ( Length of following data )

        elif data_ssl_index == 6:  # 0x0006 - hardware identification

            ssl_index_description = 'Hardware identification'

            ssl_resp_data = pack('!HHHHH20sHHH',
                                 17,  # 1  WORD   ( ID )
                                 data_ssl_index,  # 1  WORD   ( Index )
                                 28,  # 1  WORD   ( Length of payload after element count )
                                 0x01,  # 1  WORD   ( 1 element follows )
                                 data_ssl_index,  # 1  WORD   ( Data Index )
                                 str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0006'])),
                                 # 10 WORDS  ( MLFB of component: 20 bytes => 19 chars + 1 blank (0x20) )
                                 0x0,  # 1  WORD   ( RESERVED )
                                 'V3',  # 1  WORD   ( 'V' and first digit of version number )
                                 0x539)                     # 1  WORD   ( remaining digits of version number )

            ssl_resp_head = pack('!BBH',
                                 0xff,  # 1  BYTE   ( Data Error Code. 0xFF = OK )
                                 0x09,  # 1  BYTE   ( Data Type. 0x09 = Char/String )
                                 len(ssl_resp_data))  # 1  WORD   ( Length of following data )

        elif data_ssl_index == 7:  # 0x0007 - firmware identification

            ssl_index_description = 'Firmware identification'

            ssl_resp_data = pack('!HHHHH20sHHH',
                                 17,  # 1  WORD   ( ID )
                                 data_ssl_index,  # 1  WORD   ( Index )
                                 28,  # 1  WORD   ( Length of payload after element count )
                                 0x01,  # 1  WORD   ( 1 element follows )
                                 data_ssl_index,  # 1  WORD   ( Data Index )
                                 str_to_bytes(str(0x0)),  # 10 WORDS  ( RESERVED )
                                 0x0,  # 1  WORD   ( RESERVED )
                                 'V3',  # 1  WORD   ( 'V' and first digit of version number )
                                 0x53A)  # 1  WORD   ( remaining digits of version number )

            ssl_resp_head = pack('!BBH',
                                 0xff,  # 1  BYTE   ( Data Error Code. 0xFF = OK )
                                 0x09,  # 1  BYTE   ( Data Type. 0x09 = Char/String )
                                 len(ssl_resp_data))  # 1  WORD   ( Length of following data )

        else:
            ssl_index_description = 'UNKNOWN / UNDEFINED / RESERVED {0}'.format(hex(data_ssl_index))
            ssl_resp_data = ''
            ssl_resp_head = ''

        ssl_resp_params = pack('!BBBBBBBB',
                               0x00,  # SSL DIAG
                               0x01,  # unknown
                               0x12,  # unknown
                               0x08,  # bytes following
                               0x12,  # unknown, maybe 0x11 + 1
                               0x84,  # function; response to 0x44
                               0x01,  # subfunction; readszl
                               0x01)  # sequence ( = sequence + 1 )
        return ssl_index_description, ssl_resp_params, ssl_resp_head + ssl_resp_data

    # W#16#011C
    def request_ssl_28(self, data_ssl_index):

        # just for convenience
        current_ssl = S7.ssl_lists['W#16#xy1C']
        # initiate header for mass component block
        ssl_resp_data = pack('!HHHH',
                             28,  # 1  WORD   ( ID )
                             data_ssl_index,  # 1  WORD   ( Index )
                             34,  # 1  WORD   ( Length of payload after element count )
                             0x08)  # 1  WORD   ( 2 elements follow )

        # craft module data 0x0001 - automation system name
        ssl_resp_data += pack('!H24s8s',
                              0x01,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0001'])),  # TODO: PADDING
                              # 'System Name             ', # 12 WORDS  ( Name of automation system, padded with (0x00) )
                              str_to_bytes(''))  # 4  WORDS  ( RESERVED )

        # craft module data 0x0002 - component name
        ssl_resp_data += pack('!H24s8s',
                              0x02,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0002'])),
                              # 12 WORDS  ( Name of component, padded with (0x00) )
                              str_to_bytes(''))  # 4  WORDS  ( RESERVED )

        # craft module data 0x0003 - plant identification
        ssl_resp_data += pack('!H32s',
                              0x03,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0003'])),)
        # 16 WORDS  ( Name of plant, padded with (0x00) )

        # craft module data 0x0004 - copyright
        ssl_resp_data += pack('!H26s6s',
                              0x04,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0004'])),  # 13 WORDS  ( CONSTANT )
                              str_to_bytes(''))  # 3  WORDS  ( RESERVED )

        # craft module data 0x0005 - module serial number
        ssl_resp_data += pack('!H24s8s',
                              0x05,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0005'])),
                              # 12 WORDS  ( Unique Serial Number )
                              str_to_bytes(''))  # 4  WORDS  ( RESERVED )

        # craft module data 0x0007 - module type name
        ssl_resp_data += pack('!H32s',
                              0x07,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#0007'])),)
        # 16 WORDS  ( CPU type name, padded wit (0x00) )

        # craft module data 0x000a - OEM ID of module
        ssl_resp_data += pack('!H20s6s2s4s',
                              0x0a,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#000A'])),
                              # 10 WORDS  ( OEM-Copyright Text, padded with (0x00) )
                              str_to_bytes(''),  # 3  WORDS  ( OEM Copyright Text padding to 26 characters )
                              str_to_bytes(''),  # 1  WORD   ( OEM ID provided by Siemens )
                              str_to_bytes(''))  # 2  WORDS  ( OEM user defined ID )

        # craft module data 0x000b - location
        ssl_resp_data += pack('!H32s',
                              0x0b,  # 1  WORD   ( Data Index )
                              str_to_bytes(self.data_bus.get_value(current_ssl['W#16#000B'])),)
        # 16 WORDS  ( Location String, padded with (0x00) )

        # craft leading response header
        ssl_resp_head = pack('!BBH',
                             0xff,  # 1  BYTE   ( Data Error Code. 0xFF = OK )
                             0x09,  # 1  BYTE   ( Data Type. 0x09 = Char/String )
                             len(ssl_resp_data))  # 1  WORD   ( Length of following data )

        ssl_resp_packet = ssl_resp_head + ssl_resp_data
        ssl_resp_params = pack('!BBBBBBBB',
                               0x00,  # SSL DIAG
                               0x01,  # unknown
                               0x12,  # unknown
                               0x08,  # bytes following
                               0x12,  # unknown, maybe 0x11 + 1
                               0x84,  # function; response to 0x44
                               0x01,  # subfunction; readszl
                               0x01)  # sequence ( = sequence + 1 )

        return '', ssl_resp_params, ssl_resp_packet



# Copyright (C) 2014 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gevent.queue import Queue

from conpot.core.attack_session import AttackSession
from conpot.core.databus import Databus


# one instance only
class SessionManager:
    def __init__(self):
        self._sessions = []
        self._databus = Databus()
        self.log_queue = Queue()

    def _find_sessions(self, protocol, source_ip):
        for session in self._sessions:
            if session.protocol == protocol:
                if session.source_ip == source_ip:
                    return session
        return None

    def get_session(self, protocol, source_ip, source_port):
        # around here we would inject dependencies into the attack session
        attack_session = self._find_sessions(protocol, source_ip)
        if not attack_session:
            attack_session = AttackSession(protocol, source_ip, source_port, self._databus, self.log_queue)
            self._sessions.append(attack_session)
        return attack_session

    def get_session_count(self, protocol=None):
        count = 0
        if protocol:
            for session in self._sessions:
                if session.protocol == protocol:
                    count += 1
        else:
            count = len(self._sessions)
        return count

    def purge_sessions(self):
        # there is no native purge/clear mechanism for gevent queues, so...
        self.log_queue = Queue()

    def initialize_databus(self, config_file):
        self._databus.initialize(config_file)



# Copyright (C) 2015  Peter Sooky <xsooky00@stud.fit.vubtr.cz>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Author: Peter Sooky <xsooky00@stud.fit.vubtr.cz>
# Brno University of Technology, Faculty of Information Technology

import logging
import re
import sys
from bacpypes.pdu import GlobalBroadcast
import bacpypes.object
from bacpypes import errors
from bacpypes.app import BIPSimpleApplication
from bacpypes.constructeddata import Any
from bacpypes.constructeddata import InvalidParameterDatatype
from bacpypes.apdu import APDU, apdu_types, confirmed_request_types, unconfirmed_request_types, \
    ErrorPDU, RejectPDU, IAmRequest, IHaveRequest, ReadPropertyACK, ConfirmedServiceChoice, UnconfirmedServiceChoice
from bacpypes.pdu import PDU
import ast

logger = logging.getLogger(__name__)


class BACnetApp(BIPSimpleApplication):
    """
    BACnet device emulation class. BACnet properties are populated from the template file. Services are defined.
    Conpot implements a smart sensor and hence
    - DM-RP-B (execute ReadProperty)
    - DM-DDB-B (execute Who-Is, initiate I-Am)
    - DM-DOB-B (execute Who-Has, initiate I-Have)
    services are supported.
    """
    def __init__(self, device, datagram_server):
        self._request = None
        self._response = None
        self._response_service = None
        self.localDevice = device
        self.objectName = {device.objectName: device}
        self.objectIdentifier = {device.objectIdentifier: device}
        self.datagram_server = datagram_server
        self.deviceIdentifier = None
        super(BIPSimpleApplication, self).__init__()

    def get_objects_and_properties(self, dom):
        """
        parse the bacnet template for objects and their properties
        """
        self.deviceIdentifier = int(dom.xpath('//bacnet/device_info/*')[1].text)
        device_property_list = dom.xpath('//bacnet/device_info/*')
        for prop in device_property_list:
            prop_key = prop.tag.lower().title()
            prop_key = re.sub("['_','-']", "", prop_key)
            prop_key = prop_key[0].lower() + prop_key[1:]
            if prop_key not in self.localDevice.propertyList.value and \
                    prop_key not in ['deviceIdentifier', 'deviceName']:
                self.add_property(prop_key, prop.text)

        object_list = dom.xpath('//bacnet/object_list/object/@name')
        for obj in object_list:
            property_list = dom.xpath('//bacnet/object_list/object[@name="%s"]/properties/*' % obj)
            for prop in property_list:
                if prop.tag == 'object_type':
                    object_type = re.sub('-', ' ', prop.text).lower().title()
                    object_type = re.sub(' ', '', object_type) + 'Object'
            try:
                device_object = getattr(bacpypes.object, object_type)()
                device_object.propertyList = list()
            except NameError:
                logger.critical('Non-existent BACnet object type')
                sys.exit(3)
            for prop in property_list:
                prop_key = prop.tag.lower().title()
                prop_key = re.sub("['_','-']", "", prop_key)
                prop_key = prop_key[0].lower() + prop_key[1:]
                if prop_key == 'objectType':
                    prop_val = prop.text.lower().title()
                    prop_val = re.sub(" ", "", prop_val)
                    prop_val = prop_val[0].lower() + prop_val[1:]
                prop_val = prop.text
                try:
                    if prop_key == 'objectIdentifier':
                        device_object.objectIdentifier = int(prop_val)
                    else:
                        setattr(device_object, prop_key, prop_val)
                        device_object.propertyList.append(prop_key)
                except bacpypes.object.PropertyError:
                    logger.critical('Non-existent BACnet property type')
                    sys.exit(3)
            self.add_object(device_object)

    def add_object(self, obj):
        object_name = obj.objectName
        if not object_name:
            raise RuntimeError("object name required")
        object_identifier = obj.objectIdentifier
        if not object_identifier:
            raise RuntimeError("object identifier required")
        if object_name in self.objectName:
            raise RuntimeError("object already added with the same name")
        if object_identifier in self.objectIdentifier:
            raise RuntimeError("object already added with the same identifier")

        # Keep dictionaries -- for name and identifiers
        self.objectName[object_name] = obj
        self.objectIdentifier[object_identifier] = obj
        self.localDevice.objectList.append(object_identifier)

    def add_property(self, prop_name, prop_value):
        if not prop_name:
            raise RuntimeError("property name required")
        if not prop_value:
            raise RuntimeError("property value required")

        setattr(self.localDevice, prop_name, prop_value)
        self.localDevice.propertyList.append(prop_name)

    def iAm(self, *args):
        self._response = None
        return

    def iHave(self, *args):
        self._response = None
        return

    def whoIs(self, request, address, invoke_key, device):
        # Limits are optional (but if used, must be paired)
        execute = False
        try:
            if (request.deviceInstanceRangeLowLimit is not None) and \
                    (request.deviceInstanceRangeHighLimit is not None):
                if (request.deviceInstanceRangeLowLimit > list(self.objectIdentifier.keys())[0][1]
                        > request.deviceInstanceRangeHighLimit):
                    logger.info('Bacnet WhoHasRequest out of range')
                else:
                    execute = True
            else:
                execute = True
        except AttributeError:
            execute = True

        if execute:
            self._response_service = 'IAmRequest'
            self._response = IAmRequest()
            self._response.pduDestination = GlobalBroadcast()
            self._response.iAmDeviceIdentifier = self.deviceIdentifier
            # self._response.objectIdentifier = list(self.objectIdentifier.keys())[0][1]
            self._response.maxAPDULengthAccepted = int(getattr(self.localDevice, 'maxApduLengthAccepted'))
            self._response.segmentationSupported = getattr(self.localDevice, 'segmentationSupported')
            self._response.vendorID = int(getattr(self.localDevice, 'vendorIdentifier'))

    def whoHas(self, request, address, invoke_key, device):
        execute = False
        try:
            if (request.deviceInstanceRangeLowLimit is not None) and \
                    (request.deviceInstanceRangeHighLimit is not None):
                if (request.deviceInstanceRangeLowLimit > list(self.objectIdentifier.keys())[0][1]
                        > request.deviceInstanceRangeHighLimit):
                    logger.info('Bacnet WhoHasRequest out of range')
                else:
                    execute = True
            else:
                execute = True
        except AttributeError:
            execute = True

        if execute:
            for obj in device.objectList.value[2:]:
                if int(request.object.objectIdentifier[1]) == obj[1] and \
                                request.object.objectIdentifier[0] == obj[0]:
                    objName = self.objectIdentifier[obj].objectName
                    self._response_service = 'IHaveRequest'
                    self._response = IHaveRequest()
                    self._response.pduDestination = GlobalBroadcast()
                    # self._response.deviceIdentifier = list(self.objectIdentifier.keys())[0][1]
                    self._response.deviceIdentifier = self.deviceIdentifier
                    self._response.objectIdentifier = obj[1]
                    self._response.objectName = objName
                    break
            else:
                logger.info('Bacnet WhoHasRequest: no object found')

    def readProperty(self, request, address, invoke_key, device):
        # Read Property
        # TODO: add support for PropertyArrayIndex handling;
        for obj in device.objectList.value[2:]:
            if int(request.objectIdentifier[1]) == obj[1] and \
                            request.objectIdentifier[0] == obj[0]:
                objName = self.objectIdentifier[obj].objectName
                for prop in self.objectIdentifier[obj].properties:
                    if request.propertyIdentifier == prop.identifier:
                        propName = prop.identifier
                        propValue = prop.ReadProperty(
                            self.objectIdentifier[obj])
                        propType = prop.datatype()
                        self._response_service = 'ComplexAckPDU'
                        self._response = ReadPropertyACK()
                        self._response.pduDestination = address
                        self._response.apduInvokeID = invoke_key
                        self._response.objectIdentifier = obj[1]
                        self._response.objectName = objName
                        self._response.propertyIdentifier = propName

                        # get the property type
                        for p in dir(sys.modules[propType.__module__]):
                            _obj = getattr(sys.modules[propType.__module__], p)
                            try:
                                if type(propType) == _obj:
                                    break
                            except TypeError:
                                pass
                        value = ast.literal_eval(propValue)
                        self._response.propertyValue = Any(_obj(value))
                        # self._response.propertyValue.cast_in(objPropVal)
                        # self._response.debug_contents()
                        break
                else:
                    logger.info('Bacnet ReadProperty: object has no property %s', request.propertyIdentifier)
                    self._response = ErrorPDU()
                    self._response.pduDestination = address
                    self._response.apduInvokeID = invoke_key
                    self._response.apduService = 0x0c
                    # self._response.errorClass
                    # self._response.errorCode

    def indication(self, apdu, address, device):
        """logging the received PDU type and Service request"""
        request = None
        apdu_type = apdu_types.get(apdu.apduType)
        invoke_key = apdu.apduInvokeID
        logger.info('Bacnet PDU received from %s:%d. (%s)', address[0], address[1], apdu_type.__name__)
        if apdu_type.pduType == 0x0:
            # Confirmed request handling
            apdu_service = confirmed_request_types.get(apdu.apduService)
            logger.info('Bacnet indication from %s:%d. (%s)', address[0], address[1], apdu_service.__name__)
            try:
                request = apdu_service()
                request.decode(apdu)
            except (AttributeError, RuntimeError, InvalidParameterDatatype) as e:
                logger.warning('Bacnet indication: Invalid service. Error: %s' % e)
                return
            except bacpypes.errors.DecodingError:
                pass

            for key, value in list(ConfirmedServiceChoice.enumerations.items()):
                if apdu_service.serviceChoice == value:
                    try:
                        getattr(self, key)(
                            request, address, invoke_key, device)
                        break
                    except AttributeError:
                        logger.error('Not implemented Bacnet command')
                        self._response = None
                        return
            else:
                logger.info('Bacnet indication: Invalid confirmed service choice (%s)', apdu_service.__name__)
                self._response = None
                return

        # Unconfirmed request handling
        elif apdu_type.pduType == 0x1:
            apdu_service = unconfirmed_request_types.get(apdu.apduService)
            logger.info('Bacnet indication from %s:%d. (%s)', address[0], address[1], apdu_service.__name__)
            try:
                request = apdu_service()
                request.decode(apdu)
            except (AttributeError, RuntimeError):
                logger.exception('Bacnet indication: Invalid service.')
                self._response = None
                return
            except bacpypes.errors.DecodingError:
                pass

            for key, value in list(UnconfirmedServiceChoice.enumerations.items()):
                if apdu_service.serviceChoice == value:
                    try:
                        getattr(self, key)(
                            request, address, invoke_key, device)
                        break
                    except AttributeError:
                        logger.error('Not implemented Bacnet command')
                        self._response = None
                        return
            else:
                # Unrecognized services
                logger.info(
                    'Bacnet indication: Invalid unconfirmed service choice (%s)', apdu_service)
                self._response_service = 'ErrorPDU'
                self._response = ErrorPDU()
                self._response.pduDestination = address
                return
        # ignore the following
        elif apdu_type.pduType == 0x2:
            # simple ack pdu
            self._response = None
            return
        elif apdu_type.pduType == 0x3:
            # complex ack pdu
            self._response = None
            return
        elif apdu_type.pduType == 0x4:
            # segment ack
            self._response = None
            return
        elif apdu_type.pduType == 0x5:
            # error pdu
            self._response = None
            return
        elif apdu_type.pduType == 0x6:
            # reject pdu
            self._response = None
            return
        elif apdu_type.pduType == 0x7:
            # abort pdu
            self._response = None
            return
        elif 0x8 <= apdu_type.pduType <= 0xf:
            # reserved
            self._response = None
            return
        else:
            # non-BACnet PDU types
            logger.info('Bacnet Unrecognized service')
            self._response = None
            return

    # socket not actually socket, but DatagramServer with sendto method
    def response(self, response_apdu, address):
        if response_apdu is None:
            return
        apdu = APDU()
        response_apdu.encode(apdu)
        pdu = PDU()
        apdu.encode(pdu)
        if isinstance(response_apdu, RejectPDU) or isinstance(response_apdu, ErrorPDU):
            self.datagram_server.sendto(pdu.pduData, address)
        else:
            apdu_type = apdu_types.get(response_apdu.apduType)
            if pdu.pduDestination == '*:*':
                # broadcast
                # sendto operates under lock
                self.datagram_server.sendto(pdu.pduData, ('', address[1]))
            else:
                # sendto operates under lock
                self.datagram_server.sendto(pdu.pduData, address)
            logger.info('Bacnet response sent to %s (%s:%s)',
                        response_apdu.pduDestination, apdu_type.__name__, self._response_service)


# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
from gevent import select
from gevent import socket as _socket
import codecs
import gevent
from gevent.socket import socket
from gevent.ssl import wrap_socket
from gevent.server import StreamServer
import abc
import conpot.core as conpot_core


logger = logging.getLogger(__name__)


class ProxyDecoder(abc.ABC):

    @abc.abstractmethod
    def decode_in(self, data):
        """Decode data that goes into the proxied device"""

    @abc.abstractmethod
    def decode_out(self, data):
        """Decode data that goes out from the proxied device to the connected client(attacker)."""


class Proxy(object):
    def __init__(self, name, proxy_host, proxy_port, decoder=None, keyfile=None, certfile=None):
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        self.name = name
        self.proxy_id = self.name.lower().replace(' ', '_')
        self.host = None
        self.port = None
        self.keyfile = keyfile
        self.certfile = certfile
        if decoder:
            try:
                namespace, _classname = decoder.rsplit('.', 1)
                module = __import__(namespace, fromlist=[_classname])
                _class = getattr(module, _classname)
                self.decoder = _class()
                assert isinstance(self.decoder, ProxyDecoder)
            except AssertionError:
                logger.fatal('Invalid decoder: decoder must be an instance of ProxyDecoder.')
        else:
            self.decoder = None

    def get_server(self, host, port):
        self.host = host
        connection = (host, port)
        if self.keyfile and self.certfile:
            server = StreamServer(connection, self.handle, keyfile=self.keyfile, certfile=self.certfile)
        else:
            server = StreamServer(connection, self.handle)
        self.port = server.server_port
        logger.info(
            '%s proxy server started, listening on %s, proxy for: (%s, %s) using %s decoder.',
            self.name, connection, self.proxy_host, self.proxy_port, self.decoder)
        return server

    def handle(self, sock, address):
        session = conpot_core.get_session(self.proxy_id, address[0], address[1])
        logger.info(
            'New connection from %s:%s on %s proxy. (%s)',
            address[0], address[1], self.proxy_id, session.id)
        proxy_socket = socket()

        if self.keyfile and self.certfile:
            proxy_socket = wrap_socket(proxy_socket, self.keyfile, self.certfile)

        try:
            proxy_socket.connect((self.proxy_host, self.proxy_port))
        except _socket.error:
            logger.exception('Error while connecting to proxied service at ({}, {})'.format(self.proxy_host,
                             self.proxy_port))
            self._close([proxy_socket, sock])
            return

        sockets = [proxy_socket, sock]
        while len(sockets) == 2:
            gevent.sleep(0)
            sockets_read, _, sockets_err = select.select(sockets, [], sockets, 10)

            if len(sockets_err) > 0:
                self._close([proxy_socket, sock])
                break

            for s in sockets_read:
                socket_close_reason = 'socket closed'
                try:
                    data = s.recv(1024)
                except _socket.error as socket_err:
                    data = []
                    socket_close_reason = str(socket_err)
                if len(data) is 0:
                    self._close([proxy_socket, sock])
                    if s is proxy_socket:
                        logging.warning(
                            'Closing proxied socket while receiving (%s, %s): %s.',
                            self.proxy_host, self.proxy_port, socket_close_reason)
                        sockets = []
                        break
                    elif s is sock:
                        logging.warning(
                            'Closing connection to remote while receiving from remote (%s, %s): %s',
                            socket_close_reason, address[0], address[1])
                        sockets = []
                        break
                    else:
                        assert False

                try:
                    if s is proxy_socket:
                        self.handle_out_data(data, sock, session)
                    elif s is sock:
                        self.handle_in_data(data, proxy_socket, session)
                    else:
                        assert False
                except _socket.error as socket_err:
                    if s is proxy_socket:
                        destination = 'proxied socket'
                    else:
                        destination = 'remote connection'
                    logger.warning('Error while sending data to %s: %s.', destination, str(socket_err))
                    sockets = []
                    break

        session.set_ended()
        proxy_socket.close()
        sock.close()

    def handle_in_data(self, data, sock, session):
        # convert the data from bytes to hex string
        hex_data = codecs.encode(data, 'hex_codec')
        session.add_event({'raw_request': hex_data, 'raw_response': ''})
        logger.debug('Received %s bytes from outside to proxied service: %s', len(data), hex_data)
        if self.decoder:
            # TODO: data could be chunked, proxy needs to handle this
            decoded = self.decoder.decode_in(data)
            logger.debug('Decoded request: %s', decoded)
            session.add_event({'request': decoded, 'raw_response': ''})
        sock.send(data)

    def handle_out_data(self, data, sock, session):
        hex_data = codecs.encode(data, 'hex_codec')
        session.add_event({'raw_request': '', 'raw_response': hex_data})
        logger.debug('Received %s bytes from proxied service: %s', len(data), hex_data)
        if self.decoder:
            # TODO: data could be chunked, proxy needs to handle this
            decoded = self.decoder.decode_out(data)
            logger.debug('Decoded response: %s', decoded)
            session.add_event({'request': '', 'raw_response': decoded})
        sock.send(data)

    def _close(self, sockets):
        for s in sockets:
            s.close()

    def stop(self):
        # TODO: Keep active sockets in list and close them on stop()
        return


# Copyright (C) 2013  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent.monkey; gevent.monkey.patch_all()
import unittest
import datetime
from collections import namedtuple
import conpot
import os
from lxml import etree
import gevent
import requests
from gevent import socket
from conpot.protocols.http import web_server
import conpot.core as conpot_core


class TestHTTPServer(unittest.TestCase):

    def setUp(self):

        # clean up before we start...
        conpot_core.get_sessionManager().purge_sessions()
        self.dir_name = os.path.dirname(conpot.__file__)
        args = namedtuple('FakeArgs', '')
        self.http_server = web_server.HTTPServer(self.dir_name + '/templates/default/http/http.xml',
                                                 self.dir_name + '/templates/default/',
                                                 args)
        self.http_worker = gevent.spawn(self.http_server.start, '127.0.0.1', 0)
        # initialize the databus
        self.databus = conpot_core.get_databus()
        self.databus.initialize(self.dir_name + '/templates/default/template.xml')
        gevent.sleep(1)

    def tearDown(self):
        self.http_server.stop()
        gevent.joinall([self.http_worker])
        # tidy up (again)...
        conpot_core.get_sessionManager().purge_sessions()

    def test_http_request_base(self):
        """
        Objective: Test if http service delivers data on request
        """
        ret = requests.get("http://127.0.0.1:{0}/tests/unittest_base.html".format(self.http_server.server_port))
        self.assertIn('ONLINE', ret.text, "Could not retrieve expected data from test output.")

    def test_http_backend_databus(self):
        """
        Objective: Test if http backend is able to retrieve data from databus
        """
        # retrieve configuration from xml
        dom = etree.parse(self.dir_name + '/templates/default/template.xml')

        # retrieve reference value from configuration
        sysName = dom.xpath('//core/databus/key_value_mappings/key[@name="sysName"]/value')
        if sysName:
            # print(sysName)
            assert_reference = sysName[0].xpath('./text()')[0][1:-1]
        else:
            assert_reference = None
        if assert_reference is not None:
            ret = requests.get("http://127.0.0.1:{0}/tests/unittest_databus.html".format(self.http_server.server_port))
            self.assertIn(assert_reference, ret.text,
                          "Could not find databus entity 'sysName' (value '{0}') in output.".format(assert_reference))
        else:
            raise Exception("Assertion failed. Key 'sysName' not found in databus definition table.")

    def test_http_backend_tarpit(self):
        """
        Objective: Test if http tarpit delays responses properly
        """
        # retrieve configuration from xml
        dom = etree.parse(self.dir_name + '/templates/default/http/http.xml')

        # check for proper tarpit support
        tarpit = dom.xpath('//http/htdocs/node[@name="/tests/unittest_tarpit.html"]/tarpit')

        if tarpit:
            tarpit_delay = tarpit[0].xpath('./text()')[0]

            # requesting file via HTTP along with measuring the timedelta
            dt_req_start = datetime.datetime.now()
            requests.get("http://127.0.0.1:{0}/tests/unittest_tarpit.html".format(self.http_server.server_port))
            dt_req_delta = datetime.datetime.now() - dt_req_start

            # check if the request took at least the expected delay to be processed
            self.assertLessEqual(
                int(tarpit_delay),
                dt_req_delta.seconds,
                "Expected delay: >= {0} seconds. Actual delay: {1} seconds".format(tarpit_delay, dt_req_delta.seconds)
            )
        else:
            raise AssertionError("Assertion failed. Tarpit delay not found in HTTP template.")

    def test_http_subselect_trigger(self):
        """
        Objective: Test if http subselect triggers work correctly
        """
        ret = requests.get("http://127.0.0.1:{0}/tests/unittest_subselects.html?action=unit&subaction=test".format(
            self.http_server.server_port))
        self.assertIn('SUCCESSFUL', ret.text, "Trigger missed. An unexpected page was delivered.")

    def test_do_TRACE(self):
        """
        Objective: Test the web server with a trace request
        """
        # requests has no trace method.. So resorting to the good'ol socket - sending raw data
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.http_server.server_port))
        s.sendall(b'TRACE /index.html HTTP/1.1\r\nHost: localhost\r\n\r\n')
        data = s.recv(1024)
        # FIXME: Omitting the time etc from data - mechanism to check them needed as well?
        self.assertIn(b'HTTP/1.1 200 OK', data)
        # test for 501 - Disable TRACE method
        self.http_server.cmd_responder.httpd.disable_method_trace = True
        s.sendall(b'TRACE /index.html HTTP/1.1\r\nHost: localhost\r\n\r\n')
        data = s.recv(1024)
        s.close()
        self.assertIn(b'501', data)

    def test_do_HEAD(self):
        """
        Objective: Test the web server by sending a HTTP HEAD request.
        Should be responded back by the valid HTTP headers
        """
        ret = requests.head("http://127.0.0.1:{0}/tests/unittest_subselects.html?action=unit&subaction=test".format(
            self.http_server.server_port))
        self.assertTrue(ret.status_code == 200 and ret.headers['Content-Length'] == '370')

        # Test for 404
        ret = requests.head("http://127.0.0.1:{0}/tests/random_page_does_not_exists.html".format(
            self.http_server.server_port
        ))
        self.assertEqual(ret.status_code, 404)

        # test for 501 - Disable HEAD method
        self.http_server.cmd_responder.httpd.disable_method_head = True
        ret = requests.head("http://127.0.0.1:{0}/tests/unittest_subselects.html?action=unit&subaction=test".format(
            self.http_server.server_port))
        self.assertEqual(ret.status_code, 501)

    def test_do_OPTIONS(self):
        """
        Objective: Test the web server by sending a valid OPTIONS HTTP request
        """
        ret = requests.options("http://127.0.0.1:{0}/tests/unittest_subselects.html?action=unit&subaction=test".format(
            self.http_server.server_port))
        self.assertEqual((ret.headers['allow']), 'GET,HEAD,POST,OPTIONS,TRACE')
        # test for 501 - Disable OPTIONS method
        self.http_server.cmd_responder.httpd.disable_method_options = True
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(('127.0.0.1', self.http_server.server_port))
        s.sendall(b'OPTIONS /index.html HTTP/1.1\r\nHost: localhost\r\n\r\n')
        data = s.recv(1024)
        self.assertIn(b'501', data)

    def test_do_POST(self):
        """
        Objective: send a POST request to a invalid URI. Should get a 404 response
        """
        payload = {'key1': 'value1', 'key2': 'value2'}
        ret = requests.post("http://127.0.0.1:{0}/tests/demo.html".format(
            self.http_server.server_port), data=payload)
        self.assertEqual(ret.status_code, 404)

    def test_not_implemented_method(self):
        """
        Objective: PUT HTTP method is not implemented in Conpot, should raise 501
        """
        payload = b'PUT /index.html HTTP/1.1\r\nHost: localhost\r\n\r\n'
        ret = requests.put("http://127.0.0.1:{0}/tests/demo.html".format(
            self.http_server.server_port), data=payload)
        self.assertEqual(ret.status_code, 501)


if __name__ == '__main__':
    unittest.main()


# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Peter Sooky <xsooky00@stud.fit.vubtr.cz>
# Brno University of Technology, Faculty of Information Technology

import logging

from pyghmi.ipmi.bmc import Bmc


logger = logging.getLogger()


class FakeBmc(Bmc):

    def __init__(self, authdata, port):
        self.authdata = authdata
        self.port = port
        self.deviceid = 0x25
        self.revision = 0x13
        self.firmwaremajor = 0x14
        self.firmwareminor = 0x1
        self.ipmiversion = 2
        self.additionaldevices = 0
        self.mfgid = 0xf
        self.prodid = 0xe
        self.powerstate = 'off'
        self.bootdevice = 'default'
        logger.info('IPMI BMC initialized.')

    def get_boot_device(self):
        logger.info('IPMI BMC Get_Boot_Device request.')
        return self.bootdevice

    def set_boot_device(self, bootdevice):
        logger.info('IPMI BMC Set_Boot_Device request.')
        self.bootdevice = bootdevice

    def cold_reset(self):
        logger.info('IPMI BMC Cold_Reset request.')
        self.powerstate = 'off'
        self.bootdevice = 'default'

    def get_power_state(self):
        logger.info('IPMI BMC Get_Power_State request.')
        return self.powerstate

    def power_off(self):
        logger.info('IPMI BMC Power_Off request.')
        self.powerstate = 'off'

    def power_on(self):
        logger.info('IPMI BMC Power_On request.')
        self.powerstate = 'on'

    def power_reset(self):
        logger.info('IPMI BMC Power_Reset request.')
        self.powerstate = 'off'

    def power_cycle(self):
        logger.info('IPMI BMC Power_Cycle request.')
        if self.powerstate == 'off':
            self.powerstate = 'on'
        else:
            self.powerstate = 'off'

    def power_shutdown(self):
        logger.info('IPMI BMC Power_Shutdown request.')
        self.powerstate = 'off'


import os
import unittest
import conpot
import gevent
import tftpy
import filecmp
import conpot.core as conpot_core
from freezegun import freeze_time
from conpot.protocols.tftp.tftp_server import TftpServer


class TestTFTPServer(unittest.TestCase):
    def setUp(self):
        # initialize the file system.
        conpot_core.initialize_vfs()
        self._test_file = '/'.join(conpot.__path__ + ['tests/data/test_data_fs/tftp/test.txt'])
        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)
        self.tftp_server = TftpServer(self.dir_name + '/templates/default/tftp/tftp.xml',
                                      self.dir_name + '/templates/default', args=None)
        self.server_greenlet = gevent.spawn(self.tftp_server.start, '127.0.0.1', 0)
        gevent.sleep(1)

    def tearDown(self):
        self.tftp_server.stop()

    @freeze_time("2018-07-15 17:51:17")
    def test_tftp_upload(self):
        """Testing TFTP upload files. """
        client = tftpy.TftpClient('127.0.0.1', self.tftp_server.server.server_port)
        client.upload('test.txt', self._test_file)
        gevent.sleep(3)
        _, _data_fs = conpot_core.get_vfs('tftp')
        [_file] = [i for i in _data_fs.listdir('./') if '2018-07-15 17:51:17-test-txt' in i]
        self.assertEqual(_data_fs.gettext(_file), 'This is just a test file for Conpot\'s TFTP server\n')
        _data_fs.remove(_file)

    @freeze_time("2018-07-15 17:51:17")
    def test_mkdir_upload(self):
        """Testing TFTP upload files - while recursively making directories as per the TFTP path."""
        client = tftpy.TftpClient('127.0.0.1', self.tftp_server.server.server_port)
        client.upload('/dir/dir/test.txt', self._test_file)
        gevent.sleep(3)
        _, _data_fs = conpot_core.get_vfs('tftp')
        [_file] = [i for i in _data_fs.listdir('./') if '2018-07-15 17:51:17-test-txt' in i]
        self.assertEqual(_data_fs.gettext(_file), 'This is just a test file for Conpot\'s TFTP server\n')
        _data_fs.remove(_file)

    def test_tftp_download(self):
        _dst_path = '/'.join(conpot.__path__ + ['tests/data/data_temp_fs/tftp/download'])
        client = tftpy.TftpClient('127.0.0.1', self.tftp_server.server.server_port)
        try:
            client.download('tftp_data.txt', _dst_path)
            gevent.sleep(3)
            self.assertTrue(filecmp.cmp(_dst_path, self._test_file))
        finally:
            _, _data_fs = conpot_core.get_vfs('tftp')
            _data_fs.remove('download')


if __name__ == '__main__':
    unittest.main()


# Copyright (C) 2014 Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import time
import calendar


class Uptime:
    def __init__(self, started=-1):
        if started >= 0:
            initial = started
        else:
            initial = calendar.timegm(time.gmtime())
        self.started = calendar.timegm(time.gmtime(initial))

    def get_value(self):
        return calendar.timegm(time.gmtime()) - self.started



# Copyright (C) 2015 Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent
from conpot.protocols.ipmi.ipmi_server import IpmiServer
import conpot.core as conpot_core
import subprocess
from gevent import subprocess
from gevent.subprocess import PIPE
import unittest
import os
import conpot
from collections import namedtuple


def run_cmd(cmd, port):
    _cmd = ['ipmitool', '-I', 'lanplus', '-H', 'localhost', '-p', str(port), '-R1', '-U', 'Administrator', '-P',
            'Password']
    _cmd += cmd
    _process = subprocess.Popen(_cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    _result_out, _result_err = _process.communicate()
    rc = _process.returncode
    return _result_out, _result_err


class TestIPMI(unittest.TestCase):
    def setUp(self):
        # clean up before we start...
        conpot_core.get_sessionManager().purge_sessions()
        # get the current directory

        dir_name = os.path.dirname(conpot.__file__)
        args = namedtuple('FakeArgs', 'port')
        args.port = 0
        conpot_core.get_databus().initialize(dir_name + '/templates/default/template.xml')
        self.ipmi_server = IpmiServer(dir_name + '/templates/default/ipmi/ipmi.xml',
                                      dir_name + '/templates/default/', args)
        self.greenlet = gevent.spawn(self.ipmi_server.start, '127.0.0.1', 0)
        gevent.sleep(1)

    def tearDown(self):
        self.greenlet.kill()
        # tidy up (again)...
        conpot_core.get_sessionManager().purge_sessions()

    def test_boot_device(self):
        """
        Objective: test boot device get and set
        """
        result, _ = run_cmd(cmd=['chassis', 'bootdev', 'pxe'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Set Boot Device to pxe\n')

    def test_power_state(self):
        """
        Objective: test power on/off/reset/cycle/shutdown
        """
        # power status
        result, _ = run_cmd(cmd=['chassis', 'power', 'status'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Chassis Power is off\n')
        # power on
        result, _ = run_cmd(cmd=['chassis', 'power', 'on'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Chassis Power Control: Up/On\n')
        # power off
        result, _ = run_cmd(cmd=['chassis', 'power', 'off'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Chassis Power Control: Down/Off\n')
        # power reset
        result, _ = run_cmd(cmd=['chassis', 'power', 'reset'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Chassis Power Control: Reset\n')
        # power cycle
        result, _ = run_cmd(cmd=['chassis', 'power', 'cycle'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Chassis Power Control: Cycle\n')
        # shutdown gracefully
        result, _ = run_cmd(cmd=['chassis', 'power', 'soft'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Chassis Power Control: Soft\n')

    def test_chassis_status(self):
        result, _ = run_cmd(cmd=['chassis', 'status'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result,
                         b'System Power         : off\n'
                         b'Power Overload       : false\n'
                         b'Power Interlock      : inactive\n'
                         b'Main Power Fault     : false\n'
                         b'Power Control Fault  : false\n'
                         b'Power Restore Policy : always-off\n'
                         b'Last Power Event     : \n'
                         b'Chassis Intrusion    : inactive\n'
                         b'Front-Panel Lockout  : inactive\n'
                         b'Drive Fault          : false\n'
                         b'Cooling/Fan Fault    : false\n')

    def test_user_list(self):
        result, _ = run_cmd(cmd=['user', 'list'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result,
                         b'ID  Name\t     Callin  Link Auth\tIPMI Msg   Channel Priv Limit\n'
                         b'1   Administrator    true    true       true       ADMINISTRATOR\n'
                         b'2   Operator         true    false      false      OPERATOR\n'
                         b'3   User1            true    true       true       USER\n'
                         b'4   User2            true    false      false      USER\n'
                         b'5   User3            true    true       true       CALLBACK\n')

    def test_channel_get_access(self):
        result, _ = run_cmd(cmd=['channel', 'getaccess', '1', '3'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result,
                         b'Maximum User IDs     : 5\n'
                         b'Enabled User IDs     : 3\n\n'
                         b'User ID              : 3\n'
                         b'User Name            : User1\n'
                         b'Fixed Name           : Yes\n'
                         b'Access Available     : call-in / callback\n'
                         b'Link Authentication  : enabled\n'
                         b'IPMI Messaging       : enabled\n'
                         b'Privilege Level      : USER\n')

    def test_misc(self):
        # change the session pass
        result, _ = run_cmd(cmd=['set', 'password', '1', 'TEST'], port=str(self.ipmi_server.server.server_port))
        self.assertEqual(result, b'Set session password\n')


if __name__ == '__main__':
    unittest.main()